import subprocess
import sys
import os
 
# path to python.exe
python_exe = os.path.join(sys.prefix, 'bin', 'python.exe')
 
# upgrade pip
subprocess.call([python_exe, "-m", "ensurepip"])
subprocess.call([python_exe, "-m", "pip", "install", "--upgrade", "pip"])
 
# install required packages
subprocess.call([python_exe, "-m", "pip", "install", "rich"])
subprocess.call([python_exe, "-m", "pip", "install", "toml"])


# -*- coding: utf-8 -*-
"""
Created on Fri Feb 26 12:51:21 2021

@author: Rontc
"""

import numpy as np

# from mpl_toolkits.mplot3d import Axes3D
import matplotlib.pyplot as plt
import cv2
from pathlib import Path
import pyqtgraph as pg

########################################################################################################################################
########################################################################################################################################
##
##  ███    ███  █████  ████████ ██████  ██       ██████  ████████ ██      ██ ██████
##  ████  ████ ██   ██    ██    ██   ██ ██      ██    ██    ██    ██      ██ ██   ██
##  ██ ████ ██ ███████    ██    ██████  ██      ██    ██    ██    ██      ██ ██████
##  ██  ██  ██ ██   ██    ██    ██      ██      ██    ██    ██    ██      ██ ██   ██
##  ██      ██ ██   ██    ██    ██      ███████  ██████     ██    ███████ ██ ██████
##
########################################################################################################################################
########################################################################################################################################


def ReplaySkeleton_matplotlib(
    session,
    vidType=1,
    startFrame=40,
    azimuth=-90,
    elevation=-80,
    useOpenPose=True,
    useMediaPipe=False,
    useDLC=False,
):

    
    mean_charuco_fr_mar_xyz = np.load(session.dataArrayPath/'charuco_3d_points.npy')

    if useOpenPose:
        skel_fr_mar_xyz = np.load(session.dataArrayPath / "openPoseSkel_3d.npy")

    if useMediaPipe:
        mediaPipe_skel_fr_mar_xyz = np.load(
            session.dataArrayPath / "mediaPipeSkel_3d.npy"
        )

    if useDLC:
        dlc_fr_mar_xyz = np.load(session.dataArrayPath / "deepLabCut_3d.npy")
        dlc0 = np.squeeze(dlc_fr_mar_xyz[:, 0, :])
        dlc1 = np.squeeze(dlc_fr_mar_xyz[:, 1, :])
        dlc2 = np.squeeze(dlc_fr_mar_xyz[:, 2, :])
        ballTrailLen = 4

    if useOpenPose:
        for camera_number in range(session.numCams):
            imgPathList = session.session_settings['openPose_imgPathList']
  
    
    camImgPathList = {}
    for cam in range(session.numCams):
        camImgPathList[cam] = list(sorted(Path(imgPathList[cam]).glob('*.png')))
        #session.numFrames = len(camImgPathList[cam]) #will need to perhaps put a check in on whether numFrames between cameras are the same
    

# define Skeleton connections 
# head = [17, 15, 0, 16, 18]
    head = [ 15, 0, 16]
    spine = [0,1,8]
    rArm = [17, 15, 0, 16, 18]
    rArm = [4, 3, 2, 1]
    lArm = [1, 5, 6, 7]
    rLeg = [11, 10, 9, 8]
    lLeg = [14, 13, 12, 8]
    rFoot = [11, 23, 22, 11, 24]
    lFoot = [14, 20, 19, 14, 21]

    # Make some handy maps ;D
    rHandIDstart = 25
    lHandIDstart = rHandIDstart + 21

    thumb = np.array([0, 1, 2, 3, 4])
    index = np.array([0, 5, 6, 7, 8])
    bird = np.array([0, 9, 10, 11, 12])
    ring = np.array([0, 13, 14, 15, 16])
    pinky = np.array([0, 17, 18, 19, 20])

    fig = plt.figure(figsize=(11, 8))

    if vidType == 0:
        axMain = fig.add_subplot(111, projection="3d")
    elif vidType == 1:
        axMain = fig.add_subplot(121, projection="3d")
        axCam1 = fig.add_subplot(122)

    # fig.tight_layout(h_pad = 5)
    axMain.view_init(azim=azimuth, elev=elevation)
    plt.ion()
    for fr in range(startFrame, session.numFrames):
        fig.suptitle(["Frame# - ", str(fr)])
        axMain.cla()

        if vidType == 1:
            axCam1.cla()

        char_x = mean_charuco_fr_mar_xyz[:][:, 0]
        char_y = mean_charuco_fr_mar_xyz[:][:, 1]
        char_z = mean_charuco_fr_mar_xyz[:][:, 2]

        # plot skeleton (openPose by default)

        sk_x = skel_fr_mar_xyz[fr, :, 0]  # skeleton x data
        sk_y = skel_fr_mar_xyz[fr, :, 1]  # skeleton y data
        sk_z = skel_fr_mar_xyz[fr, :, 2]  # skeleton z data

        axMain.scatter3D(sk_x, sk_y, sk_z, marker=".", color="k", s=4.0)
        axMain.plot(
            sk_x[head], sk_y[head], sk_z[head], linestyle="-", color="g", linewidth=1.0
        )
        axMain.plot(
            sk_x[spine],
            sk_y[spine],
            sk_z[spine],
            linestyle="-",
            color="g",
            linewidth=1.0,
        )
        axMain.plot(
            sk_x[rArm], sk_y[rArm], sk_z[rArm], linestyle="-", color="r", linewidth=1.0
        )
        axMain.plot(
            sk_x[lArm], sk_y[lArm], sk_z[lArm], linestyle="-", color="b", linewidth=1.0
        )
        axMain.plot(
            sk_x[rLeg], sk_y[rLeg], sk_z[rLeg], linestyle="-", color="r", linewidth=1.0
        )
        axMain.plot(
            sk_x[lLeg], sk_y[lLeg], sk_z[lLeg], linestyle="-", color="b", linewidth=1.0
        )
        axMain.plot(
            sk_x[rFoot],
            sk_y[rFoot],
            sk_z[rFoot],
            linestyle="-",
            color="r",
            linewidth=1.0,
        )
        axMain.plot(
            sk_x[lFoot],
            sk_y[lFoot],
            sk_z[lFoot],
            linestyle="-",
            color="b",
            linewidth=1.0,
        )

        # plot handybois
        # right hand
        axMain.plot(
            sk_x[thumb + rHandIDstart],
            sk_y[thumb + rHandIDstart],
            sk_z[thumb + rHandIDstart],
            linestyle="-",
            color="r",
            linewidth=1.0,
        )
        axMain.plot(
            sk_x[index + rHandIDstart],
            sk_y[index + rHandIDstart],
            sk_z[index + rHandIDstart],
            linestyle="-",
            color="r",
            linewidth=1.0,
        )
        axMain.plot(
            sk_x[bird + rHandIDstart],
            sk_y[bird + rHandIDstart],
            sk_z[bird + rHandIDstart],
            linestyle="-",
            color="r",
            linewidth=1.0,
        )
        axMain.plot(
            sk_x[ring + rHandIDstart],
            sk_y[ring + rHandIDstart],
            sk_z[ring + rHandIDstart],
            linestyle="-",
            color="r",
            linewidth=1.0,
        )
        axMain.plot(
            sk_x[pinky + rHandIDstart],
            sk_y[pinky + rHandIDstart],
            sk_z[pinky + rHandIDstart],
            linestyle="-",
            color="r",
            linewidth=1.0,
        )

        # left hand
        axMain.plot(
            sk_x[thumb + lHandIDstart],
            sk_y[thumb + lHandIDstart],
            sk_z[thumb + lHandIDstart],
            linestyle="-",
            color="b",
            linewidth=1.0,
        )
        axMain.plot(
            sk_x[index + lHandIDstart],
            sk_y[index + lHandIDstart],
            sk_z[index + lHandIDstart],
            linestyle="-",
            color="b",
            linewidth=1.0,
        )
        axMain.plot(
            sk_x[bird + lHandIDstart],
            sk_y[bird + lHandIDstart],
            sk_z[bird + lHandIDstart],
            linestyle="-",
            color="b",
            linewidth=1.0,
        )
        axMain.plot(
            sk_x[ring + lHandIDstart],
            sk_y[ring + lHandIDstart],
            sk_z[ring + lHandIDstart],
            linestyle="-",
            color="b",
            linewidth=1.0,
        )
        axMain.plot(
            sk_x[pinky + lHandIDstart],
            sk_y[pinky + lHandIDstart],
            sk_z[pinky + lHandIDstart],
            linestyle="-",
            color="b",
            linewidth=1.0,
        )

        # plot mediapipe
        mp_sk_x = mediaPipe_skel_fr_mar_xyz[fr, :, 0]  # skeleton x data
        mp_sk_y = mediaPipe_skel_fr_mar_xyz[fr, :, 1]  # skeleton y data
        mp_sk_z = mediaPipe_skel_fr_mar_xyz[fr, :, 2]  # skeleton z data
        axMain.scatter3D(mp_sk_x, mp_sk_y, mp_sk_z, marker=".", color="g", s=8.0)

        # plot deeplabcut

        dlc_x = dlc_fr_mar_xyz[fr, :, 0]
        dlc_y = dlc_fr_mar_xyz[fr, :, 1]
        dlc_z = dlc_fr_mar_xyz[fr, :, 2]

        axMain.scatter3D(dlc_x, dlc_y, dlc_z, marker="o", color="r", s=24.0)

        # plot charuco grid
        axMain.scatter(char_x, char_y, char_z, marker="o")

        # #plot camera positions (I think...)
        # for camNum in range(len(session.cgroup.cameras)):
        #     axMain.scatter(session.cgroup.cameras[camNum].tvec[0], session.cgroup.cameras[camNum].tvec[1],  session.cgroup.cameras[camNum].tvec[2], marker = 'p')
        axRange = session.board.square_length * 10
        mx = np.nanmean(char_x)
        my = np.nanmean(char_y)
        mz = np.nanmean(char_z)

        axMain.set_xlim(mx - axRange, mx + axRange)
        axMain.set_ylim(my - axRange, my + axRange)
        axMain.set_zlim(mz - axRange, mz + axRange)

        axMain.set_xlabel("x")
        axMain.set_ylabel("y")
        axMain.set_zlabel("z")

        if vidType == 0:
            pass
        elif vidType == 1:
            axCam1.imshow(
                cv2.cvtColor(cv2.imread(str(camImgPathList[0][fr])), cv2.COLOR_BGR2RGB)
            )
            axCam1.axis("off")

        plt.pause(0.01)
        plt.show()

        frameName = str(fr).zfill(6)
        saveFrameName = "{}.png".format(frameName)
        saveFramePath = session.imOutPath / saveFrameName
        fig.savefig(saveFramePath)


########################################################################################################################################
########################################################################################################################################
###    ██████  ██████  ███████ ███    ██      ██████  ██
###   ██    ██ ██   ██ ██      ████   ██     ██       ██
###   ██    ██ ██████  █████   ██ ██  ██     ██   ███ ██
###   ██    ██ ██      ██      ██  ██ ██     ██    ██ ██
###    ██████  ██      ███████ ██   ████      ██████  ███████
########################################################################################################################################
########################################################################################################################################



def ReplaySkeleton_matplotlib( session,
                    vidType=1,
                    startFrame=40,
                    azimuth=-90,
                    elevation=-80,
                    useOpenPose=True,
                    useMediaPipe=False,
                    useDLC=False):

    session.imOutPath.mkdir(exist_ok = True)
    mean_charuco_fr_mar_xyz = np.load(session.dataArrayPath/'charuco_3d_points.npy')


    if session.useMediaPipe:
        mediaPipe_skel_fr_mar_xyz = np.load(session.dataArrayPath/'mediaPipeSkel_3d.npy')
        imgPathList = session.session_settings['mediaPipe_imgPathList']
        #session.numCams = len(session.mediaPipe_imgPathList)

    if session.useOpenPose:
        skel_fr_mar_xyz = np.load(session.dataArrayPath/'openPoseSkel_3d.npy')
        imgPathList = session.session_settings['openPose_imgPathList']
        #session.numCams = len(session.openPose_imgPathList)    
        
    if session.useDLC:
        dlc_fr_mar_xyz = np.load(session.dataArrayPath / "deepLabCut_3d.npy")
        dlc0 = np.squeeze(dlc_fr_mar_xyz[:, 0, :])
        dlc1 = np.squeeze(dlc_fr_mar_xyz[:, 1, :])
        dlc2 = np.squeeze(dlc_fr_mar_xyz[:, 2, :])
        ballTrailLen = 4

    camImgPathList = {}
    for cam in range(session.numCams):
        camImgPathList[cam] = list(sorted(Path(imgPathList[cam]).glob('*.png')))
        #session.numFrames = len(camImgPathList[cam]) #will need to perhaps put a check in on whether numFrames between cameras are the same
    

# define Skeleton connections 
# head = [17, 15, 0, 16, 18]
    head = [ 15, 0, 16]
    spine = [0,1,8]
    rArm = [4 ,3 ,2 ,1]
    lArm = [1, 5, 6, 7]
    rLeg = [11, 10, 9, 8]
    lLeg = [14, 13, 12, 8]
    rFoot = [11, 23, 22, 11, 24]
    lFoot = [14, 20, 19, 14, 21]

    # Make some handy maps ;D
    rHandIDstart = 25
    lHandIDstart = rHandIDstart + 21

    thumb = np.array([0, 1, 2, 3, 4])
    index = np.array([0, 5, 6, 7, 8])
    bird = np.array([0, 9, 10, 11, 12])
    ring = np.array([0, 13, 14, 15, 16])
    pinky = np.array([0, 17, 18, 19, 20])

    fig = plt.figure(figsize=(11, 8))

    if vidType == 0:
        axMain = fig.add_subplot(111, projection="3d")
    elif vidType == 1:
        axMain = fig.add_subplot(121, projection="3d")
        axCam1 = fig.add_subplot(122)

    # fig.tight_layout(h_pad = 5)
    axMain.view_init(azim=azimuth, elev=elevation)
    plt.ion()
    for fr in range(startFrame, session.numFrames):
        fig.suptitle(["Frame# - ", str(fr)])
        axMain.cla()

        if vidType == 1:
            axCam1.cla()

        char_x = mean_charuco_fr_mar_xyz[:][:, 0]
        char_y = mean_charuco_fr_mar_xyz[:][:, 1]
        char_z = mean_charuco_fr_mar_xyz[:][:, 2]

        # plot skeleton (openPose by default)
        if session.useOpenPose:
            sk_x = skel_fr_mar_xyz[fr, :, 0]  # skeleton x data
            sk_y = skel_fr_mar_xyz[fr, :, 1]  # skeleton y data
            sk_z = skel_fr_mar_xyz[fr, :, 2]  # skeleton z data

            axMain.scatter3D(sk_x, sk_y, sk_z, marker=".", color="k", s=4.0)
            axMain.plot(
                sk_x[head],
                sk_y[head],
                sk_z[head],
                linestyle="-",
                color="g",
                linewidth=1.0,
            )
            axMain.plot(
                sk_x[spine],
                sk_y[spine],
                sk_z[spine],
                linestyle="-",
                color="g",
                linewidth=1.0,
            )
            axMain.plot(
                sk_x[rArm],
                sk_y[rArm],
                sk_z[rArm],
                linestyle="-",
                color="r",
                linewidth=1.0,
            )
            axMain.plot(
                sk_x[lArm],
                sk_y[lArm],
                sk_z[lArm],
                linestyle="-",
                color="b",
                linewidth=1.0,
            )
            axMain.plot(
                sk_x[rLeg],
                sk_y[rLeg],
                sk_z[rLeg],
                linestyle="-",
                color="r",
                linewidth=1.0,
            )
            axMain.plot(
                sk_x[lLeg],
                sk_y[lLeg],
                sk_z[lLeg],
                linestyle="-",
                color="b",
                linewidth=1.0,
            )
            axMain.plot(
                sk_x[rFoot],
                sk_y[rFoot],
                sk_z[rFoot],
                linestyle="-",
                color="r",
                linewidth=1.0,
            )
            axMain.plot(
                sk_x[lFoot],
                sk_y[lFoot],
                sk_z[lFoot],
                linestyle="-",
                color="b",
                linewidth=1.0,
            )

            # plot handybois
            # right hand
            axMain.plot(
                sk_x[thumb + rHandIDstart],
                sk_y[thumb + rHandIDstart],
                sk_z[thumb + rHandIDstart],
                linestyle="-",
                color="r",
                linewidth=1.0,
            )
            axMain.plot(
                sk_x[index + rHandIDstart],
                sk_y[index + rHandIDstart],
                sk_z[index + rHandIDstart],
                linestyle="-",
                color="r",
                linewidth=1.0,
            )
            axMain.plot(
                sk_x[bird + rHandIDstart],
                sk_y[bird + rHandIDstart],
                sk_z[bird + rHandIDstart],
                linestyle="-",
                color="r",
                linewidth=1.0,
            )
            axMain.plot(
                sk_x[ring + rHandIDstart],
                sk_y[ring + rHandIDstart],
                sk_z[ring + rHandIDstart],
                linestyle="-",
                color="r",
                linewidth=1.0,
            )
            axMain.plot(
                sk_x[pinky + rHandIDstart],
                sk_y[pinky + rHandIDstart],
                sk_z[pinky + rHandIDstart],
                linestyle="-",
                color="r",
                linewidth=1.0,
            )

            # left hand
            axMain.plot(
                sk_x[thumb + lHandIDstart],
                sk_y[thumb + lHandIDstart],
                sk_z[thumb + lHandIDstart],
                linestyle="-",
                color="b",
                linewidth=1.0,
            )
            axMain.plot(
                sk_x[index + lHandIDstart],
                sk_y[index + lHandIDstart],
                sk_z[index + lHandIDstart],
                linestyle="-",
                color="b",
                linewidth=1.0,
            )
            axMain.plot(
                sk_x[bird + lHandIDstart],
                sk_y[bird + lHandIDstart],
                sk_z[bird + lHandIDstart],
                linestyle="-",
                color="b",
                linewidth=1.0,
            )
            axMain.plot(
                sk_x[ring + lHandIDstart],
                sk_y[ring + lHandIDstart],
                sk_z[ring + lHandIDstart],
                linestyle="-",
                color="b",
                linewidth=1.0,
            )
            axMain.plot(
                sk_x[pinky + lHandIDstart],
                sk_y[pinky + lHandIDstart],
                sk_z[pinky + lHandIDstart],
                linestyle="-",
                color="b",
                linewidth=1.0,
            )

        # plot mediapipe
        if session.useMediaPipe:
            mp_sk_x = mediaPipe_skel_fr_mar_xyz[fr, :, 0]  # skeleton x data
            mp_sk_y = mediaPipe_skel_fr_mar_xyz[fr, :, 1]  # skeleton y data
            mp_sk_z = mediaPipe_skel_fr_mar_xyz[fr, :, 2]  # skeleton z data
            axMain.scatter3D(mp_sk_x, mp_sk_y, mp_sk_z, marker=".", color="g", s=8.0)

        # plot deeplabcut
        if session.useDLC:
            dlc_x = dlc_fr_mar_xyz[fr, :, 0]
            dlc_y = dlc_fr_mar_xyz[fr, :, 1]
            dlc_z = dlc_fr_mar_xyz[fr, :, 2]

            axMain.scatter3D(dlc_x, dlc_y, dlc_z, marker="o", color="r", s=24.0)

        # plot charuco grid
        axMain.scatter(char_x, char_y, char_z, marker="o")

        # #plot camera positions (I think...)
        # for camNum in range(len(session.cgroup.cameras)):
        #     axMain.scatter(session.cgroup.cameras[camNum].tvec[0], session.cgroup.cameras[camNum].tvec[1],  session.cgroup.cameras[camNum].tvec[2], marker = 'p')
        axRange = session.board.square_length * 10
        mx = np.nanmean(char_x)
        my = np.nanmean(char_y)
        mz = np.nanmean(char_z)

        axMain.set_xlim(mx - axRange, mx + axRange)
        axMain.set_ylim(my - axRange, my + axRange)
        axMain.set_zlim(mz - axRange, mz + axRange)

        axMain.set_xlabel("x")
        axMain.set_ylabel("y")
        axMain.set_zlabel("z")

        if vidType == 0:
            pass
        elif vidType == 1:
            axCam1.imshow(
                cv2.cvtColor(cv2.imread(str(camImgPathList[0][fr])), cv2.COLOR_BGR2RGB)
            )
            axCam1.axis("off")

        plt.pause(0.01)
        plt.show()

        frameName = str(fr).zfill(6)
        saveFrameName = "{}.png".format(frameName)
        saveFramePath = session.imOutPath / saveFrameName
        fig.savefig(saveFramePath)



from src.cameras.detection.cam_detection import DetectPossibleCameras
from src.cameras.detection.models import FoundCamerasResponse

# No consumer should call this "private" variable

_available_cameras: FoundCamerasResponse = None


# If you want cams, you call this function
def get_or_create_cams():
    global _available_cameras
    if _available_cameras is None:
        d = DetectPossibleCameras()
        _available_cameras = d.find_available_cameras()

    return _available_cameras



# define point parents by name. Only points listed will have a parent. If a point is not listed, it will be without parents
# child_name : parent_name
parent_mapping = {
    'REye': 'Nose',
    'LKnee': 'LHip',
    'LWrist': 'LElbow',
    'LSmallToe': 'LBigToe',
    'RShoulder': 'Neck',
    'LBigToe': 'LAnkle',
    'RKnee': 'RHip',
    'LEar': 'LEye',
    'REar': 'REye',
    'RBigToe': 'RAnkle',
    'RAnkle': 'RKnee',
    'LElbow': 'LShoulder',
    'RElbow': 'RShoulder',
    'LAnkle': 'LKnee',
    'RHeel': 'RAnkle',
    'RWrist': 'RElbow',
    'RSmallToe': 'RBigToe',
    'RHip': 'MidHip',
    'Nose': 'Neck',
    'MidHip': 'Neck',
    'LShoulder': 'Neck',
    'LEye': 'Nose',
    'LHip': 'MidHip',
    'LHeel': 'LAnkle',
    'HandL15': 'HandL14',
    #'HandL13': 'HandL00',
    'HandL14': 'HandL13',
    'HandL20': 'HandL19',
    'RWrist': 'RElbow',
    'HandL12': 'HandL11',
    'HandL18': 'HandL17',
    #'HandL02': 'HandL01',
    'HandL03': 'HandL02',
    #'HandL17': 'HandL00',
    #'HandL01': 'HandL00',
    'HandL06': 'HandL05',
    'HandL07': 'HandL06',
    'HandL04': 'HandL03',
    #'HandL05': 'HandL00',
    'HandL11': 'HandL10',
    'HandL08': 'HandL07',
    #'HandL09': 'HandL00',
    'HandL19': 'HandL18',
    'HandL16': 'HandL15',
    'HandL10': 'HandL09',
    'HandR14': 'HandR13',
    'HandR08': 'HandR07',
    #'HandR09': 'HandR00',
    'HandR19': 'HandR18',
    'HandR18': 'HandR17',
    'HandR04': 'HandR03',
    #'HandR05': 'HandR00',
    'HandR06': 'HandR05',
    'HandR07': 'HandR06',
    #'HandR13': 'HandR00',
    #'HandR01': 'HandR00',
    #'HandR02': 'HandR01',
    'HandR03': 'HandR02',
    #'HandR17': 'HandR00',
    'HandR20': 'HandR19',
    'LWrist': 'LElbow',
    'HandR11': 'HandR10',
    'HandR12': 'HandR11',
    'HandR16': 'HandR15',
    'HandR10': 'HandR09',
    'HandR15': 'HandR14'
}


# point name corresponding to Open Pose list indices
# If the imported list is longer than the point_name_mapping, all residual points will be given a generic name
point_name_mapping = [
    "Nose", "Neck", "RShoulder", "RElbow", "RWrist", "LShoulder",               # limbs
    "LElbow", "LWrist", "MidHip", "RHip", "RKnee", "RAnkle", "LHip", "LKnee",   # limbs
    "LAnkle", "REye", "LEye", "REar", "LEar", "LBigToe", "LSmallToe", "LHeel",  # limbs
    "RBigToe", "RSmallToe", "RHeel",                                            # limbs
    "HandR00", "HandR01", "HandR02", "HandR03", "HandR04", "HandR05", "HandR06", "HandR07", "HandR08", "HandR09", "HandR10", "HandR11", "HandR12", "HandR13", "HandR14", "HandR15", "HandR16", "HandR17", "HandR18", "HandR19", "HandR20", # Right Hand
    "HandL00", "HandL01", "HandL02", "HandL03", "HandL04", "HandL05", "HandL06", "HandL07", "HandL08", "HandL09", "HandL10", "HandL11", "HandL12", "HandL13", "HandL14", "HandL15", "HandL16", "HandL17", "HandL18", "HandL19", "HandL20", # Left Hand
]



from pyqtgraph.Qt import QtCore, QtGui
import pyqtgraph.opengl as gl
import pyqtgraph as pg
import pyqtgraph.console
from pyqtgraph.dockarea import Dock, DockArea
import cv2

import numpy as np
from pathlib import Path
import sys

 
class PlaySkeletonWidget:
    def __init__(self, session):

        # set up PyQT widget (the window that pops up and shows the plots, I think ? cribbed from - https://gist.github.com/markjay4k/da2f55e28514be7160a7c5fbf95bd243)
        # self.app = QtGui.QApplication(sys.argv)
        self.Skel3dViewWidget = gl.GLViewWidget()
        self.Skel3dViewWidget.opts["distance"] = 2000
        self.Skel3dViewWidget.setWindowTitle("SessionID: {}".format(session.sessionID))
        self.Skel3dViewWidget.setGeometry(0, 110, 1920, 1080)
        self.Skel3dViewWidget.show()

        # load data to be plotted
        self.useOpenPose = session.useOpenPose
        if self.useOpenPose:
            self.skel_fr_mar_xyz = np.load(
                session.dataArrayPath / "openPoseSkel_3d.npy"
            )
            self.skelScatterItem = gl.GLScatterPlotItem(
                pos=self.skel_fr_mar_xyz, color=(1, 0, 1, 0.8), size=10, pxMode=False
            )

            self.Skel3dViewWidget.addItem(self.skelScatterItem)

        elif session.useMediaPipe:
            self.skel_fr_mar_xyz = np.load(
                session.dataArrayPath / "mediaPipeSkel_3d.npy"
            )
            self.skelScatterItem = gl.GLScatterPlotItem(
                pos=self.skel_fr_mar_xyz, color=(1, 0, 1, 0.8), size=10, pxMode=False
            )

            self.Skel3dViewWidget.addItem(self.skelScatterItem)

            # make skeleton line plots (numbers specify what order to connects the dots)
        #     self.head = [17, 15, 0, 1, 0, 16, 18]
        #     self.spine = [1, 8]
        #     self.rArm = [17, 15, 0, 16, 18]
        #     self.rArm = [4, 3, 2, 1]
        #     self.lArm = [1, 5, 6, 7]
        #     self.rLeg = [11, 10, 9, 8]
        #     self.lLeg = [14, 13, 12, 8]
        #     self.rFoot = [11, 23, 22, 11, 24]
        #     self.lFoot = [14, 20, 19, 14, 21]

        #     # Make some handy maps ;D
        #     rHandIDstart = 25
        #     lHandIDstart = rHandIDstart + 21

        #     self.rHand = (
        #         np.array(
        #             [
        #                 0,
        #                 1,
        #                 2,
        #                 3,
        #                 4,
        #                 False,
        #                 0,
        #                 5,
        #                 6,
        #                 7,
        #                 8,
        #                 False,
        #                 0,
        #                 9,
        #                 10,
        #                 11,
        #                 12,
        #                 False,
        #                 0,
        #                 13,
        #                 14,
        #                 15,
        #                 16,
        #                 False,
        #                 0,
        #                 17,
        #                 18,
        #                 19,
        #                 20,
        #             ]
        #         )
        #         + rHandIDstart
        #     )
        #     self.lHand = self.rHand + lHandIDstart

        #     self.headLineItem = gl.GLLinePlotItem(
        #         pos=self.skel_fr_mar_xyz[:, self.head, :],
        #         color=(1, 1, 1, 0.2),
        #         width=4.0,
        #         antialias=True,
        #     )
        #     self.Skel3dViewWidget.addItem(self.headLineItem)

        #     self.spineLineItem = gl.GLLinePlotItem(
        #         pos=self.skel_fr_mar_xyz[:, self.spine, :],
        #         color=(1, 1, 1, 0.9),
        #         width=4.0,
        #         antialias=True,
        #     )
        #     self.Skel3dViewWidget.addItem(self.spineLineItem)

        #     self.rArmLineItem = gl.GLLinePlotItem(
        #         pos=self.skel_fr_mar_xyz[:, self.rArm, :],
        #         color=(1, 0.2, 0.15, 0.9),
        #         width=4.0,
        #         antialias=True,
        #     )
        #     self.Skel3dViewWidget.addItem(self.rArmLineItem)

        #     self.lArmLineItem = gl.GLLinePlotItem(
        #         pos=self.skel_fr_mar_xyz[:, self.lArm, :],
        #         color=(0.4, 0.6, 1, 0.9),
        #         width=4.0,
        #         antialias=True,
        #     )
        #     self.Skel3dViewWidget.addItem(self.lArmLineItem)

        #     self.rLegLineItem = gl.GLLinePlotItem(
        #         pos=self.skel_fr_mar_xyz[:, self.rLeg, :],
        #         color=(1, 0.2, 0.15, 0.9),
        #         width=4.0,
        #         antialias=True,
        #     )
        #     self.Skel3dViewWidget.addItem(self.rLegLineItem)

        #     self.lLegLineItem = gl.GLLinePlotItem(
        #         pos=self.skel_fr_mar_xyz[:, self.lLeg, :],
        #         color=(0.4, 0.6, 1, 0.9),
        #         width=4.0,
        #         antialias=True,
        #     )
        #     self.Skel3dViewWidget.addItem(self.lLegLineItem)

        #     self.rFootLineItem = gl.GLLinePlotItem(
        #         pos=self.skel_fr_mar_xyz[:, self.rFoot, :],
        #         color=(1, 0.2, 0.15, 0.9),
        #         width=4.0,
        #         antialias=True,
        #     )
        #     self.Skel3dViewWidget.addItem(self.rFootLineItem)

        #     self.lFootLineItem = gl.GLLinePlotItem(
        #         pos=self.skel_fr_mar_xyz[:, self.lFoot, :],
        #         color=(0.4, 0.6, 1, 0.9),
        #         width=4.0,
        #         antialias=True,
        #     )
        #     self.Skel3dViewWidget.addItem(self.lFootLineItem)

        #     # plot handybois
        #     self.rHandLineItem = gl.GLLinePlotItem(
        #         pos=self.skel_fr_mar_xyz[:, self.rHand, :],
        #         color=(0.2, 0.4, 1, 0.9),
        #         width=1.0,
        #         antialias=True,
        #     )
        #     # self.Skel3dViewWidget.addItem(self.rHandLineItem)

        #     self.lHandLineItem = gl.GLLinePlotItem(
        #         pos=self.skel_fr_mar_xyz[:, self.lHand, :],
        #         color=(1, 0.1, 0.05, 0.9),
        #         width=1.0,
        #         antialias=True,
        #     )
        #     # self.Skel3dViewWidget.addItem(self.lHandLineItem)

        self.useDLC = session.useDLC
        if self.useDLC:
            # self.dlc_fr_mar_xyz = np.load(session.dataArrayPath/'mediaPipeSkel_3d.npy')
            self.dlc_fr_mar_xyz = np.load(session.dataArrayPath / "deepLabCut_3d.npy")
            self.dlcScatterItem = gl.GLScatterPlotItem(
                pos=self.dlc_fr_mar_xyz, color=(1, 0, 0, 0.8), size=60, pxMode=False
            )
            self.Skel3dViewWidget.addItem(self.dlcScatterItem)

            self.dlc0LineItem = gl.GLLinePlotItem(
                pos=self.dlc_fr_mar_xyz[:, 0, :],
                color=(0, 1, 1, 0.5),
                width=4.0,
                antialias=True,
            )
            self.dlc1LineItem = gl.GLLinePlotItem(
                pos=self.dlc_fr_mar_xyz[:, 1, :],
                color=(1, 0, 1, 0.5),
                width=4.0,
                antialias=True,
            )
            self.dlc2LineItem = gl.GLLinePlotItem(
                pos=self.dlc_fr_mar_xyz[:, 2, :],
                color=(1, 1, 0, 0.5),
                width=4.0,
                antialias=True,
            )

            self.Skel3dViewWidget.addItem(self.dlc0LineItem)
            self.Skel3dViewWidget.addItem(self.dlc1LineItem)
            self.Skel3dViewWidget.addItem(self.dlc2LineItem)

            self.tailLength = 10

        self.charuco_mar_dim = np.load(session.dataArrayPath/'charuco_3d_points.npy')
        self.charucoScatterItem = gl.GLScatterPlotItem(pos=self.charuco_mar_dim, color=(1,1,1,.8), size=10, pxMode=False)
        self.Skel3dViewWidget.addItem(self.charucoScatterItem)

        # plot videos Frames!

        # set frame counter
        self.currentFrameNumber = 0
        self.framerate = 50  # frames per second
        self.frameDuration = 1000 / self.framerate  # milliseconds

    # def start(self):
    #     if (sys.flags.interactive != 1) or not hasattr(QtCore, 'PYQT_VERSION'):
    #         QtGui.QApplication.instance().exec_()

    def update(self):

        # Plots up Mr. Skreleton :D
        # Plots the dots
        
        self.skelScatterItem.setData(
            pos=self.skel_fr_mar_xyz[self.currentFrameNumber, :, :]
        )

        # connect the dottos
        if self.useOpenPose:
            self.headLineItem.setData(
                pos=self.skel_fr_mar_xyz[self.currentFrameNumber, self.head, :]
            )
            self.spineLineItem.setData(
                pos=self.skel_fr_mar_xyz[self.currentFrameNumber, self.spine, :]
            )
            self.rArmLineItem.setData(
                pos=self.skel_fr_mar_xyz[self.currentFrameNumber, self.rArm, :]
            )
            self.lArmLineItem.setData(
                pos=self.skel_fr_mar_xyz[self.currentFrameNumber, self.lArm, :]
            )
            self.rLegLineItem.setData(
                pos=self.skel_fr_mar_xyz[self.currentFrameNumber, self.rLeg, :]
            )
            self.lLegLineItem.setData(
                pos=self.skel_fr_mar_xyz[self.currentFrameNumber, self.lLeg, :]
            )
            self.rFootLineItem.setData(
                pos=self.skel_fr_mar_xyz[self.currentFrameNumber, self.rFoot, :]
            )
            self.lFootLineItem.setData(
                pos=self.skel_fr_mar_xyz[self.currentFrameNumber, self.lFoot, :]
            )

            self.rHandLineItem.setData(
                pos=self.skel_fr_mar_xyz[self.currentFrameNumber, self.rHand, :]
            )
            self.lHandLineItem.setData(
                pos=self.skel_fr_mar_xyz[self.currentFrameNumber, self.lHand, :]
            )

        if self.useDLC:
            # Plot DeepLabCut data
            self.dlcScatterItem.setData(
                pos=self.dlc_fr_mar_xyz[self.currentFrameNumber, :, :]
            )

            if self.currentFrameNumber > self.tailLength:
                self.tailIndexes = range(
                    self.currentFrameNumber - self.tailLength, self.currentFrameNumber
                )
                self.dlc0LineItem.setData(
                    pos=self.dlc_fr_mar_xyz[self.tailIndexes, 0, :]
                )
                self.dlc1LineItem.setData(
                    pos=self.dlc_fr_mar_xyz[self.tailIndexes, 1, :]
                )
                self.dlc2LineItem.setData(
                    pos=self.dlc_fr_mar_xyz[self.tailIndexes, 2, :]
                )

        self.currentFrameNumber += 1
        if self.currentFrameNumber >= self.skel_fr_mar_xyz.shape[0]:
            self.currentFrameNumber = 0

    def animate(self):
        timer = QtCore.QTimer()
        timer.timeout.connect(self.update)
        # timer.start(self.frameDuration)
        timer.start(0)
        # self.start()


class VideoWindowWidget:
    def __init__(self, vidNum, videoPath, session):
        # set up video capture
        self.vidCap = cv2.VideoCapture(str(videoPath))
        success, image = self.vidCap.read()
        self.vidCap.set(
            cv2.CAP_PROP_POS_FRAMES, 0
        )  # reset video capture so things start on the first frame

        self.videoWidget = pg.GraphicsLayoutWidget()
        self.videoWidget.show()  ## show widget alone in its own window
        self.videoWidget.setWindowTitle("Video Name")
        self.view = self.videoWidget.addViewBox()
        ## lock the aspect ratio so pixels are always square
        self.view.setAspectLocked(True)

        ## Create image item
        self.imgItem = pg.ImageItem(border="w")
        self.view.addItem(self.imgItem)

        ## Set initial view bounds
        self.view.setRange(QtCore.QRectF(0, 0, image.shape[0], image.shape[1]))

        # set image, I guess?
        self.imgItem.setImage(image)

    def update(self):
        success, image = self.vidCap.read()

        if success:
            image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
            image = cv2.rotate(image, cv2.ROTATE_90_CLOCKWISE)
            self.imgItem.setImage(image)
        else:
            self.vidCap.set(
                cv2.CAP_PROP_POS_FRAMES, 0
            )  # reset video capture so video will loop
            success, image = self.vidCap.read()
            self.imgItem.setImage(image)


class PlayerDockedWindow:
    def __init__(self, session, displayVid):

        app = pg.mkQApp("DockArea Example")
        self.displayVid = displayVid
        self.win = QtGui.QMainWindow()
        area = DockArea()
        self.win.setCentralWidget(area)
        winHeight = 1000
        winWidth = 1900
        self.win.resize(winHeight, winWidth)
        self.win.setWindowTitle("FreeMoCap")

        ## Create docks, place them into the window one at a time.
        ## Note that size arguments are only a suggestion; docks will still have to
        ## fill the entire dock area and obey the limits of their internal widgets.

        self.VidPathList = []
        self.dock_name_dictionary = {}

        if displayVid == 0:
            path_to_search = session.syncedVidPath
        elif displayVid == 1:
            path_to_search = session.openPoseDataPath
        for count, vidPath in enumerate(path_to_search.glob("*.mp4")):
            self.VidPathList.append(vidPath)
            # dock_name = 'dock_video{}'.format(count)

            dock_name = count
            self.dock_name_dictionary[dock_name] = None

        dock_3dView = Dock(
            "Session: {}".format(session.sessionID), size=(1, 1), closable=True
        )  ## give this dock the minimum possible size
        dock_Console = Dock(
            "Dock2 - Console", size=(winHeight, winWidth), closable=True
        )

        for key in self.dock_name_dictionary:
            self.dock_name_dictionary[key] = Dock(str(key), size=(1, 1), closable=True)
        # dock_video0 = Dock("Video0", size=(1,1), closable=True)
        # dock_video1 = Dock("Video1", size=(1,1), closable=True)
        # dock_video2 = Dock("Video2", size=(1,1), closable=True)
        # dock_video3 = Dock("Video3", size=(1,1), closable=True)

        area.addDock(
            dock_3dView, "left"
        )  ## place d1 at left edge of dock area (it will fill the whole space since there are no other docks yet)
        direction_list = ["top", "left", "bottom", "top"]
        for key, value in self.dock_name_dictionary.items():
            if (key % 2) == 0:
                area.addDock(value, "right")
                previous_dock = value
            else:
                area.addDock(value, "bottom", previous_dock)
                # previous_dock = value

        # area.addDock(dock_video0, 'right')     ## place d2 at right edge of dock area
        # area.addDock(dock_video1, 'right', dock_video0)## place d3 at bottom edge of d1
        # area.addDock(dock_video2, 'bottom',dock_video0)     ## place d4 at right edge of dock area
        # area.addDock(dock_video3, 'left', dock_video2)  ## place d5 at left edge of d1
        # area.moveDock(dock_video1, 'right', dock_video0)

        ## Add widgets into each dock

        ## first dock gets save/restore buttons
        w1 = pg.LayoutWidget()
        label = QtGui.QLabel(
            """ -- DockArea Example -- 
        This window has 6 Dock widgets in it. Each dock can be dragged
        by its title bar to occupy a different space within the window 
        but note that one dock has its title bar hidden). Additionally,
        the borders between docks may be dragged to resize. Docks that are dragged on top
        of one another are stacked in a tabbed layout. Double-click a dock title
        bar to place it in its own window.
        """
        )

        w2 = pg.console.ConsoleWidget()
        dock_Console.addWidget(w2)

        self.playSkel = PlaySkeletonWidget(session)

        dock_3dView.addWidget(self.playSkel.Skel3dViewWidget)

        # self.VidPathList = []
        # for vidPath in session.syncedVidPath.glob('*.mp4'):
        #    self.VidPathList.append(vidPath)
        self.widget_list = []
        for key, dock_video in self.dock_name_dictionary.items():
            self.videoWidget = VideoWindowWidget(
                key, str(self.VidPathList[key]), session
            )
            self.widget_list.append(self.videoWidget)
            dock_video.addWidget(self.videoWidget.videoWidget)
        # self.video0Widget = VideoWindowWidget(0,str(self.VidPathList[0]), session)
        # dock_video0.addWidget(self.video0Widget.videoWidget )

        # self.video1Widget = VideoWindowWidget(1,str(self.VidPathList[1]),session)
        # dock_video1.addWidget(self.video1Widget.videoWidget )

        # self.video2Widget = VideoWindowWidget(2,str(self.VidPathList[2]), session)
        # dock_video2.addWidget(self.video2Widget.videoWidget )

        # self.video3Widget = VideoWindowWidget(3,str(self.VidPathList[3]),session)
        # dock_video3.addWidget(self.video3Widget.videoWidget )

        self.win.show()

    def update(self):
        self.playSkel.update()
        for widget in self.widget_list:
            widget.update()

        # self.video0Widget.update()
        # self.video1Widget.update()
        # self.video2Widget.update()
        # self.video3Widget.update()

    def animate(self):
        timer = QtCore.QTimer()
        timer.timeout.connect(self.update)
        timer.start(self.playSkel.frameDuration)
        self.start()

    def start(self):
        if (sys.flags.interactive != 1) or not hasattr(QtCore, "PYQT_VERSION"):
            QtGui.QApplication.instance().exec_()



import logging

import cv2

from src.cameras.capture.dataclasses.frame_payload import FramePayload
from src.core_processor.camera_calibration.charuco_board_detection.dataclasses.charuco_board_definition import CharucoBoard
from src.core_processor.camera_calibration.charuco_board_detection.dataclasses.charuco_frame_payload import \
    CharucoFramePayload
from src.core_processor.camera_calibration.charuco_board_detection.dataclasses.charuco_view_data import  CharucoViewData

from src.core_processor.camera_calibration.charuco_board_detection.charuco_image_annotator import (
    annotate_image_with_charuco_data,
)

logger = logging.getLogger(__name__)


class CharucoBoardDetector:
    def __init__(self):
        charuco_board_object = CharucoBoard()
        self.charuco_board = charuco_board_object.charuco_board
        self.aruco_marker_dict = charuco_board_object.aruco_marker_dict
        self.number_of_charuco_corners = charuco_board_object.number_of_charuco_corners

    def detect_charuco_board(self, raw_frame_payload: FramePayload) -> CharucoFramePayload:

        if raw_frame_payload.image is None:
            logger.error("Frame is empty")
            return CharucoFramePayload()

        charuco_view_data = self.detect_charuco_board_in_an_image(raw_frame_payload.image)

        annotated_image = raw_frame_payload.image.copy()
        if charuco_view_data.some_charuco_corners_found:
            annotate_image_with_charuco_data(
                annotated_image,  # this image will have stuff drawn on top of it inside this function
                charuco_view_data,
                self.number_of_charuco_corners
            )

        return CharucoFramePayload(
            annotated_image=annotated_image,
            raw_frame_payload=raw_frame_payload,
            charuco_view_data=charuco_view_data
        )

    def detect_charuco_board_in_an_image(self, raw_image) -> CharucoViewData:
        """
        detect charuco board in an image.
        more-or-less copied from - https://mecaruco2.readthedocs.io/en/latest/notebooks_rst/Aruco/sandbox/ludovic/aruco_calibration_rotation.html

        """
        if raw_image is None:
            logger.error("Image is empty!")
            return CharucoViewData()

        image_height, image_width, num_color_channels = raw_image.shape
        charuco_corners = []
        charuco_ids = []

        # SUB PIXEL CORNER DETECTION CRITERION
        termination_criteria_threshold = (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, 100, 0.00001)

        grayscale_image = cv2.cvtColor(raw_image, cv2.COLOR_BGR2GRAY)

        aruco_square_corners, aruco_square_ids, rejected_image_points = cv2.aruco.detectMarkers(
            grayscale_image, self.aruco_marker_dict
        )

        full_board_found = False
        any_markers_found = False
        some_charuco_corners_found = False

        if len(aruco_square_corners) > 0:
            any_markers_found = True
            # refine detected corner locations to provide sub-pixel precision
            # https://docs.opencv.org/4.x/dd/d1a/group__imgproc__feature.html#ga354e0d7c86d0d9da75de9b9701a9a87e
            for this_corner in aruco_square_corners:
                cv2.cornerSubPix(
                    grayscale_image, this_corner, winSize=(3, 3), zeroZone=(-1, -1),
                    criteria=termination_criteria_threshold
                )
            results = cv2.aruco.interpolateCornersCharuco(aruco_square_corners,
                                                          aruco_square_ids,
                                                          grayscale_image,
                                                          self.charuco_board
                                                          )

            if results[1] is not None and results[2] is not None and len(results[1]) > 3:
                some_charuco_corners_found = True

                charuco_corners = results[1]
                charuco_ids = results[2]

                if len(charuco_ids) == self.number_of_charuco_corners:
                    full_board_found = True

        return CharucoViewData(full_board_found=full_board_found,
                               some_charuco_corners_found=some_charuco_corners_found,
                               any_markers_found=any_markers_found,
                               charuco_corners=charuco_corners,
                               charuco_ids=charuco_ids,
                               aruco_square_corners=aruco_square_corners,
                               aruco_square_ids=aruco_square_ids,
                               image_width=image_width,
                               image_height=image_height)


if __name__ == "__main__":
    pass
    # ## Jon, Run me to easily start a charuco board detect run
    # print('starting main')
    # BoardDetection().process()



from typing import List

import cv2
import numpy as np

from src.core_processor.camera_calibration.charuco_board_detection.dataclasses.charuco_view_data import CharucoViewData


def _board_text(
        charuco_ids: List[int],
        num_charuco_corners: int,
        full_board_detected_bool: bool,
):
    num_ids_as_str = str(len(charuco_ids))
    num_corners_as_str = str(num_charuco_corners)
    text = (
        f"{num_ids_as_str} of {num_corners_as_str} Charuco Corner Points Detected "
        f"| Full Board Detected: {full_board_detected_bool}"
    )
    return text


def annotate_image_with_charuco_data(
        image,
        charuco_view_data: CharucoViewData,
        number_of_charuco_corners: int) -> bool:

    annotated_image = cv2.aruco.drawDetectedCornersCharuco(
        image,
        np.array(charuco_view_data.charuco_corners),
        np.array(charuco_view_data.charuco_ids),
        (0, 255, 125, 255)
    )  # yellow? I
    # think cv2 uses BGR instead of RGB?


    current_cam_corner_count_str = _board_text(
        charuco_view_data.charuco_ids,
        number_of_charuco_corners,
        charuco_view_data.full_board_found,
    )
    # TODO - Determine 'shared views' (i.e. frames in which a full board is detected by 2
    #  cameras)
    # TODO - self.determine_shared_charuco_board_views()
    # this_cam_shared_views_str = " | Shared Views: " + str(
    #     each_cameras_shared_board_view_count_total)
    text_to_write_on_this_camera = current_cam_corner_count_str

    position = (10, 50)
    cv2.putText(
        annotated_image,  # numpy array on which text is written
        text_to_write_on_this_camera,  # text
        position,  # position at which writing has to start
        cv2.FONT_HERSHEY_SIMPLEX,  # font family
        0.5,  # font size
        (30, 10, 0, 255),  # font color
        2,
    )  # font stroke (draw a darker heavier font beneath a lighter/thinner copy for
    # readability)

    cv2.putText(
        annotated_image,  # numpy array on which text is written
        text_to_write_on_this_camera,  # text
        position,  # position at which writing has to start
        cv2.FONT_HERSHEY_SIMPLEX,
        # font family (very limited selection, i think there's some interesting CV history
        # here...)
        0.5,  # font size
        (209, 180, 0, 255),  # font color
        1,
    )  # font stroke

    if charuco_view_data.full_board_found:
        cv2.polylines(
            annotated_image, np.int32([charuco_view_data.charuco_corners]), False, (0, 100, 255), 2
        )



"""Packages the addon and dependency"""
import importlib
import os
import re
import zipfile

import freemocap_blender_addon

ignored_directories = (
    '__pycache__',
    '.DS_Store'
)


def gather_files(basedir, arc_prefix=''):
    """Walk the given directory and return a sequence of filepath, archive name
    pairs.

    Args:
        basedir: The directory to start the walk

        arc_prefix: A path to join to the front of the relative (to basedir)
            file path

    Returns:
        A sequence of (filepath, archive name) pairs
    """
    results = []

    for path, subdirectories, files in os.walk(basedir):
        if os.path.basename(path) in ignored_directories:
            continue

        for file in files:
            relative_path = os.path.join(path, file)
            full_path = os.path.abspath(relative_path)
            arcname = os.path.relpath(full_path, os.path.dirname(basedir))
            arcname = os.path.join(arc_prefix, arcname)

            results.append((full_path, arcname))

    return results


def get_required_modules():
    """Parse the requirements.txt file to determine dependencies.

    Returns:
        A sequence of module names
    """
    with open('requirements.txt') as file:
        data = file.read()
        modules = data.split('\n')
        pattern = '([A-Za-z0-9]+)(?:[<=>]+.*\n)?'

        def get_module_name(s):
            result = re.search(pattern, s)

            if result:
                return result.group()

        modules = [get_module_name(s) for s in modules if s]

        return modules


def run():
    try:
        os.mkdir('dist')

    except FileExistsError:
        pass

    zip_entries = gather_files('freemocap_blender_addon')

    for module in get_required_modules():
        module = importlib.import_module(module)
        zip_entries += gather_files(os.path.dirname(module.__file__), os.path.join('freemocap_blender_addon', 'modules'))

    filename = f'freemocap_blender_addon-{freemocap_blender_addon.__version__}.zip'
    filepath = os.path.abspath(os.path.join('dist', filename))
    with zipfile.ZipFile(filepath, 'w') as dist_zip:
        for filename, arcname in zip_entries:
            dist_zip.write(filename, arcname)


if __name__ == '__main__':
    run()



import logging
from dataclasses import dataclass

import cv2
import numpy as np

from src.cameras.capture.dataclasses.frame_payload import FramePayload
from src.cameras.capture.opencv_camera.opencv_camera import OpenCVCamera
from src.core_processor.camera_calibration.calibration_dataclasses import CameraCalibrationData
from src.core_processor.camera_calibration.charuco_board_detection.charuco_board_detector import CharucoBoardDetector
from src.core_processor.camera_calibration.charuco_board_detection.charuco_image_annotator import \
    annotate_image_with_charuco_data
from src.core_processor.camera_calibration.charuco_board_detection.dataclasses.charuco_view_data import CharucoViewData

logger = logging.getLogger(__name__)


class CameraCalibrator:
    def __init__(self,
                 charuco_board_detector: CharucoBoardDetector = CharucoBoardDetector(),
                 ):
        self._current_calibration = None
        self._charuco_board_detector = charuco_board_detector

    def calibrate(self, this_camera: OpenCVCamera):
        annotated_image = self._process_incoming_frame(this_camera.latest_frame)

        return annotated_image

    def _process_incoming_frame(self, raw_frame_payload: FramePayload):

        charuco_frame_payload = self._charuco_board_detector.detect_charuco_board(raw_frame_payload)

        charuco_view_data = charuco_frame_payload.charuco_view_data

        if self._current_calibration is None:
            self._current_calibration = CameraCalibrationData(charuco_view_data.image_width,
                                                              charuco_view_data.image_height)

        if not charuco_view_data.full_board_found and self._current_calibration is None:
            return charuco_frame_payload.annotated_image, self._current_calibration


        if charuco_view_data.full_board_found:
            self._calibrate_camera(charuco_view_data)  # this is where the magic happens


        if charuco_view_data.some_charuco_corners_found:
            annotate_image_with_charuco_data(
                raw_frame_payload.image,  # this image will have stuff drawn on top of it inside this function
                charuco_view_data,
                self._charuco_board_detector.number_of_charuco_corners
            )

        undistorted_image = cv2.undistort(raw_frame_payload.image,
                                          self._current_calibration.camera_matrix,
                                          self._current_calibration.lens_distortion_coefficients,
                                          )

        return undistorted_image, self._current_calibration

    def _calibrate_camera(self, charuco_view: CharucoViewData):
        """
        adapted from - https://mecaruco2.readthedocs.io/en/latest/notebooks_rst/Aruco/sandbox/ludovic/aruco_calibration_rotation.html

        Calibrates the camera using the charuco data

        helpful resources -
        https://docs.opencv.org/4.x/d9/d0c/group__calib3d.html <-opencv's 3d camera calibration docs
        https://www.adamsmith.haus/python/docs/cv2.aruco.calibrateCameraCharucoExtended
        https://www.mathworks.com/help/vision/ug/camera-calibration.html
        https://docs.opencv.org/3.4/d9/db7/tutorial_py_table_of_contents_calib3d.html
        https://calib.io/blogs/knowledge-base/camera-models <-a good one
        """
        print("CAMERA CALIBRATION")

        image_width = charuco_view.image_width
        image_height = charuco_view.image_height


        # flag definitions -> https://docs.opencv.org/4.x/d9/d0c/group__calib3d.html#ga3207604e4b1a1758aa66acb6ed5aa65d
        flags = (cv2.CALIB_USE_INTRINSIC_GUESS +
                 cv2.CALIB_ZERO_TANGENT_DIST +
                 cv2.cv2.CALIB_FIX_ASPECT_RATIO +
                 cv2.CALIB_FIX_PRINCIPAL_POINT +
                 # cv2.CALIB_FIX_FOCAL_LENGTH
                 cv2.CALIB_RATIONAL_MODEL
                 # cv2.CALIB_THIN_PRISM_MODEL
                 # cv2.CALIB_TILTED_MODEL
                 )

        # https://docs.opencv.org/4.x/d9/d6a/group__aruco.html#gacf03e5afb0bc516b73028cf209984a06
        (this_reprojection_error,
         this_camera_matrix,
         these_lens_distortion_coefficients,
         these_rotation_vectors_of_the_board,
         these_translation_vectors_of_the_board,
         this_lens_distortion_std_dev,
         this_camera_location_std_dev,
         these_reprojection_error_per_view
         ) = cv2.aruco.calibrateCameraCharucoExtended(
            charucoCorners=[charuco_view.charuco_corners],
            charucoIds=[charuco_view.charuco_ids],
            board=self._charuco_board_detector.charuco_board,
            imageSize=(image_width, image_height),
            cameraMatrix=self._current_calibration.camera_matrix,
            distCoeffs=self._current_calibration.lens_distortion_coefficients,
            flags=flags,
            criteria=(cv2.TERM_CRITERIA_EPS & cv2.TERM_CRITERIA_COUNT, 10000, 1e-9))

        if this_reprojection_error < self._current_calibration.reprojection_error:
            self._current_calibration = CameraCalibrationData(
                reprojection_error=this_reprojection_error,
                camera_matrix=this_camera_matrix,
                lens_distortion_coefficients=these_lens_distortion_coefficients,
                image_width=image_width,
                image_height=image_height,
                rotation_vectors_of_the_board=these_rotation_vectors_of_the_board,
                translation_vectors_of_the_board=these_translation_vectors_of_the_board,
                lens_distortion_std_dev=this_lens_distortion_std_dev,
                camera_location_std_dev=this_camera_location_std_dev)



import logging
import time
from typing import Dict

import cv2
import numpy as np
import pyqtgraph as pg
import pyqtgraph.opengl as gl
from pyqtgraph.dockarea.Dock import Dock
from pyqtgraph.dockarea.DockArea import DockArea
from pyqtgraph.Qt import QtWidgets

from src.core_processor.fps.timestamp_manager import TimestampManager

logger = logging.getLogger(__name__)


class QTVisualizerAndGui:
    def __init__(self):
        # https://pyqtgraph.readthedocs.io/en/latest/config_options.html
        self._camera0_id = '0'
        pg.setConfigOptions(imageAxisOrder='row-major')
        self.pyqtgraph_app = pg.mkQApp("FreeMoCap! :O ")

        self._dict_of_camera_image_item_widgets = {}
        self._dict_of_simple_timestamp_line_plots = {}
        self._dict_of_cam_timestamps = {}
        self._dict_of_timestamp_difference_line_plots = {}
        self._dict_of_cam_timestamp_differences = {}
        self._dict_of_timestamp_difference_histograms = {}
        self._number_of_cameras = None
        self._webcam_ids_list = []

        self._is_paused = False

    @property
    def pause_button_pressed(self):
        return self._is_paused

    def close(self):
        self._main_window_widget.close()

    def setup_and_launch(self, webcam_ids_list):
        logger.info('setting up QT Visualizer and GUI')
        self._number_of_cameras = len(webcam_ids_list)
        self._webcam_ids_list = webcam_ids_list
        self._setup_main_window()
        self._setup_control_panel()
        self._setup_camera_views_dock()
        self._setup_timestamp_plot()
        self._setup_time_difference_from_cam0_line_plot()
        self._setup_time_difference_from_cam0_histogram_plot()
        self._setup_3d_viewport()
        logger.info('launching QT Visualizer and GUI window')
        self._main_window_widget.show()

    def update_camera_view_image(self, webcam_id, image_to_display):
        if self.pause_button_pressed:
            return

        try:
            camera_image_item_widget = self._dict_of_camera_image_item_widgets[webcam_id]
        except Exception as e:
            logger.warning(f'Could not find ViewBoxWidget for camera {webcam_id}')
            raise e
        camera_image_item_widget.setImage(cv2.cvtColor(image_to_display, cv2.COLOR_BGR2RGB))

    def update_timestamp_plots(self, timestamp_manager: TimestampManager):
        if self.pause_button_pressed:
            return
        if timestamp_manager.get_timestamps_from_camera(self._camera0_id).number_of_frames <= 0:
            return

        self._update_simple_timestamp_plot(timestamp_manager)
        self._update_timestamp_difference_plot(timestamp_manager)
        self._update_timestamp_difference_histogram(timestamp_manager)

    def _update_simple_timestamp_plot(self, timestamp_manager: TimestampManager):
        for webcam_id in self._webcam_ids_list:
            plot_item = self._dict_of_simple_timestamp_line_plots[webcam_id]
            plot_item.setData(timestamp_manager.get_timestamps_from_camera(webcam_id).timestamps)

    def _update_timestamp_difference_plot(self, timestamp_manager: TimestampManager):
        for webcam_id in self._webcam_ids_list:
            plot_item = self._dict_of_simple_timestamp_line_plots[webcam_id]
            plot_item.setData(np.diff(timestamp_manager.get_timestamps_from_camera(webcam_id).timestamps))

    def _update_timestamp_difference_histogram(self, timestamp_manager: TimestampManager):
        for webcam_id in self._webcam_ids_list:
            this_cam_timestamp_diffs = np.diff(timestamp_manager.get_timestamps_from_camera(webcam_id).timestamps)
            num_samples = this_cam_timestamp_diffs.shape[0]
            counts, bin_edges = np.histogram(this_cam_timestamp_diffs,
                                             bins=np.linspace(0, 100, 100))
            counts = counts / num_samples
            try:
                self._dict_of_timestamp_difference_histograms[webcam_id].setData(
                    x=bin_edges,
                    y=counts)
            except Exception as e:
                logger.warning(f'Could not find timestamp difference line plot widget for camera {webcam_id}')
                raise e

    def _setup_main_window(self, window_width: int = 1000, window_height: int = 1000):
        """
        This is the main main window for the GUI.
        Its structure is based loosely on the 'Dock Widgets' example from `python -m pyqtgraph.examples`
        """
        self._main_window_widget = QtWidgets.QMainWindow()
        self._main_window_widget.resize(window_width, window_height)

        self._main_dock_area = DockArea()
        self._main_window_widget.setCentralWidget(self._main_dock_area)
        self._main_window_widget.setWindowTitle('Main Window ;D')

    def _setup_control_panel(self):
        self._control_panel_dock = Dock('Control Panel', size=(1, 1))
        control_panel_layout_widget = pg.LayoutWidget()
        self._control_panel_dock.addWidget(control_panel_layout_widget)

        label = QtWidgets.QLabel("Blah Blah Blah put words here ")
        control_panel_layout_widget.addWidget(label, row=0, col=0)

        self._pause_button = QtWidgets.QPushButton('Pause')
        self._pause_button.setEnabled(True)
        self._pause_button.clicked.connect(self._pause)
        control_panel_layout_widget.addWidget(self._pause_button, row=1, col=0)

        self._play_button = QtWidgets.QPushButton('Resume')
        self._play_button.setEnabled(False)
        self._play_button.clicked.connect(self._play)
        control_panel_layout_widget.addWidget(self._play_button, row=2, col=0)

        self._reset_calibration_button = QtWidgets.QPushButton('Reset Calibration')
        self._reset_calibration_button.setEnabled(False)
        self._reset_calibration_button.clicked.connect(self._reset_calibration)
        control_panel_layout_widget.addWidget(self._reset_calibration_button, row=3, col=0)

        self._record_button = QtWidgets.QPushButton('Record')
        self._record_button.setEnabled(False)
        self._record_button.clicked.connect(self._record)
        control_panel_layout_widget.addWidget(self._record_button, row=3, col=0)

        self._main_dock_area.addDock(self._control_panel_dock, position='left')

    def _setup_camera_views_dock(self):
        self._camera_graphics_layout_window = pg.GraphicsLayoutWidget()
        self._camera_views_dock = Dock('Camera Views')
        self._camera_views_dock.addWidget(self._camera_graphics_layout_window)
        self._main_dock_area.addDock(self._camera_views_dock, 'right', self._control_panel_dock)

        for this_webcam_id in self._webcam_ids_list:
            self._dict_of_camera_image_item_widgets[this_webcam_id] = self._create_camera_view_widget()

    def _create_camera_view_widget(self):
        camera_view_box_widget = pg.ViewBox(invertY=True, lockAspect=True)
        camera_image_item = pg.ImageItem()
        camera_view_box_widget.addItem(camera_image_item)
        self._camera_graphics_layout_window.addItem(camera_view_box_widget)
        return camera_image_item

    def _setup_timestamp_plot(self):
        # create widget/dock for reprojection error subplot
        timestamp_plot_widget = pg.PlotWidget(title="Timestamp vs Frame#")
        timestamp_plot_widget.setLabel('left', "timestamp", units='seconds')
        timestamp_plot_widget.setLabel('bottom', "Frame#", units='Frame')
        timestamp_plot_widget.addLegend()

        for this_cam_num, this_webcam_id in enumerate(self._webcam_ids_list):
            this_timestamp_plot_line = timestamp_plot_widget.plot(np.empty(0),
                                                                  pen=(this_cam_num, self._number_of_cameras),
                                                                  name="camera " + this_webcam_id)

            self._dict_of_simple_timestamp_line_plots[this_webcam_id] = this_timestamp_plot_line
            self._dict_of_cam_timestamps[this_webcam_id] = np.ndarray(0)

        self._timestamp_plot_dock = Dock("Camera Timestamps")
        self._timestamp_plot_dock.addWidget(timestamp_plot_widget)
        self._main_dock_area.addDock(self._timestamp_plot_dock, 'bottom')

    def _setup_time_difference_from_cam0_line_plot(self):
        # create widget/dock for reprojection error subplot
        timestamp_difference_plot_widget = pg.PlotWidget(title="Timestamp difference from Camera0 on each frame")
        timestamp_difference_plot_widget.setLabel('left', "np.abs(this_camera timestamp - cam0 timestamp)",
                                                  units='milliseconds')
        timestamp_difference_plot_widget.setLabel('bottom', "Frame#", units='Frame')
        timestamp_difference_plot_widget.addLegend()

        for this_cam_num, this_webcam_id in enumerate(self._webcam_ids_list):
            this_plot_line = timestamp_difference_plot_widget.plot(np.empty(0),
                                                                   pen=(this_cam_num, self._number_of_cameras),
                                                                   name="camera " + this_webcam_id)

            self._dict_of_timestamp_difference_line_plots[this_webcam_id] = this_plot_line
            self._dict_of_cam_timestamp_differences[this_webcam_id] = np.ndarray(0)

        self._timestamp_difference_plot_dock = Dock("Camera Timestamps")
        self._timestamp_difference_plot_dock.addWidget(timestamp_difference_plot_widget)
        self._main_dock_area.addDock(self._timestamp_difference_plot_dock, 'right', self._timestamp_plot_dock)

    def _setup_time_difference_from_cam0_histogram_plot(self):
        timestamp_difference_histogram_plot_widget = pg.PlotWidget(
            title="Timestamp difference from Camera0 on each frame")
        timestamp_difference_histogram_plot_widget.setLabel('left', "Proportion")
        timestamp_difference_histogram_plot_widget.setLabel('bottom', "milliseconds")
        timestamp_difference_histogram_plot_widget.addLegend()

        for this_cam_num, this_webcam_id in enumerate(self._webcam_ids_list):
            this_histogram_plot_item = timestamp_difference_histogram_plot_widget.plot(np.empty(0),
                                                                                       np.empty(0),
                                                                                       stepMode="center",
                                                                                       fillLevel=0,
                                                                                       fillOutline=False,
                                                                                       brush=pg.mkBrush(color=(
                                                                                           this_cam_num,
                                                                                           self._number_of_cameras),
                                                                                           alpha=.5),
                                                                                       name="camera " + this_webcam_id)

            self._dict_of_timestamp_difference_histograms[this_webcam_id] = this_histogram_plot_item

        self._timestamp_diff_histgoram_dock = Dock("Camera Timestamp Difference histogram")
        self._timestamp_diff_histgoram_dock .addWidget(timestamp_difference_histogram_plot_widget)
        self._main_dock_area.addDock(self._timestamp_diff_histgoram_dock, 'right', self._timestamp_difference_plot_dock)

    def _pause(self):
        self._is_paused = True
        self._play_button.setEnabled(True)
        self._play_button.setText("GUI Paused - Main loop still active")
        self._pause_button.setEnabled(False)

    def _play(self):
        self._is_paused = False
        self._play_button.setEnabled(False)
        self._pause_button.setEnabled(True)

    def _reset_calibration(self):
        pass

    def _record(self):
        pass

    def _setup_3d_viewport(self):
        self.opengl_3d_plot_widget = gl.GLViewWidget()
        self.opengl_grid_item = gl.GLGridItem()
        self.opengl_3d_plot_widget.addItem(self.opengl_grid_item)

        # create XYZ axes
        x_axis_line_array = np.array([[0, 0, 0], [1, 0, 0]])
        y_axis_line_array = np.array([[0, 0, 0], [0, 1, 0]])
        z_axis_line_array = np.array([[0, 0, 0], [0, 0, 1]])

        self.origin_x_axis_gl_lineplot_item = pg.opengl.GLLinePlotItem(pos=x_axis_line_array,
                                                                       color=(1, 0, 0, 1),
                                                                       width=1.,
                                                                       antialias=True)
        self.origin_y_axis_gl_lineplot_item = pg.opengl.GLLinePlotItem(pos=y_axis_line_array,
                                                                       color=(0, 1, 0, 1),
                                                                       width=1.,
                                                                       antialias=True)
        self.origin_z_axis_gl_lineplot_item = pg.opengl.GLLinePlotItem(pos=z_axis_line_array,
                                                                       color=(0, 0, 1, 1),
                                                                       width=1.,
                                                                       antialias=True)
        self.opengl_3d_plot_widget.addItem(self.origin_x_axis_gl_lineplot_item)
        self.opengl_3d_plot_widget.addItem(self.origin_y_axis_gl_lineplot_item)
        self.opengl_3d_plot_widget.addItem(self.origin_z_axis_gl_lineplot_item)

        self.opengl_charuco_scatter_item = gl.GLScatterPlotItem(pos=(0, 0, 0), color=(1, 0, 1), size=1,
                                                                       pxMode=False)

        self.opengl_3d_plot_dock = Dock("3d View Port")
        self.opengl_3d_plot_dock.addWidget(self.opengl_3d_plot_widget)
        self._main_dock_area.addDock(self.opengl_3d_plot_dock, 'right', self._timestamp_diff_histgoram_dock )



import time
import logging
import traceback
from pathlib import Path

import cv2

from src.cameras.multicam_manager.cv_camera_manager import OpenCVCameraManager
from src.cameras.persistence.video_writer.save_options_dataclass import SaveOptions
from src.config.data_paths import freemocap_data_path
from src.core_processor.camera_calibration.camera_calibrator import CameraCalibrator
from src.core_processor.fps.timestamp_manager import TimestampManager
from src.core_processor.mediapipe_skeleton_detection.mediapipe_skeleton_detection import MediaPipeSkeletonDetection
from src.core_processor.show_cam_window import show_cam_window
from src.qt_visualizer_and_gui.qt_visualizer_and_gui import QTVisualizerAndGui

logger = logging.getLogger(__name__)

def get_canonical_time_str():
    return time.strftime("%m-%d-%Y-%H_%M_%S")


class SessionPipelineOrchestrator:
    def __init__(self):
        self._session_start_time_unix_ns = time.time_ns()
        self._session_id = 'session_' + time.strftime("%m-%d-%Y-%H_%M_%S")
        self._visualizer_gui = QTVisualizerAndGui()
        self._open_cv_camera_manager = OpenCVCameraManager(session_id=self._session_id)
        self._camera_calibrator = CameraCalibrator()
        self._mediapipe_skeleton_detector = MediaPipeSkeletonDetection(model_complexity=2)

    @property
    def session_folder_path(self):
        return freemocap_data_path / self._session_id

    async def process_by_cam_id(self, webcam_id: str, cb):
        with self._open_cv_camera_manager.start_capture_session_single_cam(
                webcam_id=webcam_id,
        ) as session_obj:
            fps_manager = TimestampManager(self._open_cv_camera_manager.available_webcam_ids)
            fps_manager.start_all()
            cv_cam = session_obj.cv_camera
            writer = session_obj.video_recorder
            try:
                while True:
                    if not cv_cam.is_capturing_frames:
                        logger.debug(f'Camera {webcam_id} is not capturing frames')
                        continue

                    if self.calibrate_cameras:
                        charuco_frame_payload = self._camera_calibrator.calibrate(cv_cam)

                    if cb and charuco_frame_payload.annotated_image is not None:
                        writer.record(charuco_frame_payload.raw_frame_payload)
                        await cb(charuco_frame_payload.annotated_image)
            except Exception as e:
                logger.error("Printing traceback")
                traceback.print_exc()
                raise e
            finally:
                writer = session_obj.video_recorder
                options = SaveOptions(
                    writer_dir=Path().joinpath(
                        cv_cam.session_writer_base_path,
                        "charuco_board_detection",
                        f"webcam_{cv_cam.webcam_id_as_str}",
                    ),
                    fps=fps_manager.median_frames_per_second(cv_cam.webcam_id_as_str),
                    frame_width=cv_cam.image_width(),
                    frame_height=cv_cam.image_height(),
                )
                writer.save_to_disk(options)

    def run(
            self,
            show_camera_views_in_windows=True,
            show_visualizer_gui=True,
            calibrate_cameras=True,
            detect_skeleton=True,
            save_video=True,
    ):
        """
        Opens Cameras using OpenCV and begins image processing for charuco board
        If return images is true, the images are returned to the caller
        """
        with self._open_cv_camera_manager.start_capture_session_all_cams() as connected_cameras_dict:
            try:
                if show_visualizer_gui:
                    self._visualizer_gui.setup_and_launch(self._open_cv_camera_manager.available_webcam_ids)

                timestamp_manager = TimestampManager(self._open_cv_camera_manager.available_webcam_ids)

                should_continue = True
                while should_continue:

                    for this_webcam_id, this_open_cv_camera in connected_cameras_dict.items():

                        if not this_open_cv_camera.new_frame_ready:
                            continue

                        this_cam_latest_frame = this_open_cv_camera.latest_frame

                        if this_cam_latest_frame is None:
                            continue

                        image_to_display = this_cam_latest_frame.image
                        this_cam_this_frame_timestamp = this_cam_latest_frame.timestamp

                        timestamp_manager.increment_frame_processed_for(this_webcam_id,
                                                                        this_cam_this_frame_timestamp)

                        if save_video:
                            this_open_cv_camera.video_recorder.record(this_cam_latest_frame)

                        if calibrate_cameras:
                            undistorted_annotated_image = self._camera_calibrator.calibrate(this_open_cv_camera)
                            image_to_display = undistorted_annotated_image

                        if detect_skeleton:
                            image_to_display = self._mediapipe_skeleton_detector.detect_skeleton_in_image(
                                image_to_display)

                        if show_camera_views_in_windows:
                            should_continue = show_cam_window(
                                this_webcam_id, image_to_display, timestamp_manager
                            )

                        if show_visualizer_gui:
                            self._visualizer_gui.update_camera_view_image(this_webcam_id, image_to_display)
                            # self._visualizer_gui.update_timestamp_plots(timestamp_manager)

            except:
                logger.error("Printing traceback")
                traceback.print_exc()
            finally:
                for this_open_cv_camera in connected_cameras_dict.values():
                    if save_video:
                        this_open_cv_camera.video_recorder.save_to_disk(
                            self.session_folder_path / 'synchronized_videos')
                    logger.info(f"Destroy window {this_open_cv_camera.webcam_id_as_str}")
                    cv2.destroyWindow(this_open_cv_camera.webcam_id_as_str)
                    cv2.waitKey(1)

                if show_visualizer_gui:
                    self._visualizer_gui.close()


if __name__ == "__main__":
    print('start main')

    this_session = SessionPipelineOrchestrator()
    this_session.run(save_video=False,
                     show_visualizer_gui=True,
                     calibrate_cameras=False,
                     detect_skeleton=True)



from src.api.routes.camera.cam_data_ws import cam_ws_router
from src.api.routes.camera.camera_route import camera_router
from src.api.routes.health.health_check_route import healthcheck_router
from src.api.routes.startup.startup import startup_router

# REGISTER NEW ROUTES HERE
enabled_routers = [healthcheck_router, camera_router, cam_ws_router, startup_router]



from dataclasses import dataclass

import numpy as np

@dataclass
class CharucoViewData:
    charuco_corners: np.ndarray = None
    charuco_ids: np.ndarray = None
    aruco_square_corners: np.ndarray = None
    aruco_square_ids: np.ndarray = None
    full_board_found: bool = False
    some_charuco_corners_found: bool = False
    any_markers_found: bool = False
    image_width: int = None
    image_height: int = None





from typing import NamedTuple, Union

import numpy as np


class FramePayload(NamedTuple):
    success: bool = False
    image: np.ndarray = None
    timestamp: Union[int, float] = None
    webcam_id: str = None





import logging
import platform

import cv2

from src.cameras.detection.models import FoundCamerasResponse, RawCamera

CAM_CHECK_NUM = 20

logger = logging.getLogger(__name__)


class DetectPossibleCameras:
    def find_available_cameras(self) -> FoundCamerasResponse:
        cv2_backend = self._determine_backend()

        cams_to_use_list = []
        for cam_id in range(CAM_CHECK_NUM):
            cap = cv2.VideoCapture(cam_id, cv2_backend)
            success, image = cap.read()

            if success:
                try:
                    cams_to_use_list.append(
                        RawCamera(
                            webcam_id=str(cam_id)
                        )
                    )
                finally:
                    cap.release()

        return FoundCamerasResponse(
            number_of_cameras_found=len(cams_to_use_list),
            cameras_found_list=cams_to_use_list,
            cv2_backend=cv2_backend,
        )

    def _determine_backend(self):
        if platform.system() == "Windows":
            return cv2.CAP_DSHOW
        else:
            return cv2.CAP_ANY



import numpy as np
import bpy

from pathlib import Path
import json
from mathutils import Vector

from rich import print, inspect
from rich.progress import track
from rich.console import Console

console = Console()

console.rule()
console.rule()
console.rule("Hello! getting started with FreeMoCap Loading Stuff :D")
console.rule()

# load mediapipe 3d skeleton data

sessionID = "sesh_2021-12-30_13_57_13_cleanish"

freemocap_data_path = Path("C:/Users/jonma/Dropbox/FreeMoCapProject/FreeMocap_Data")
session_path = freemocap_data_path / sessionID
data_path = session_path / "DataArrays"
mediapipe_skel_data_path = data_path / "mediaPipeSkel_3d_smoothed.npy"

console.rule(style="color({})".format("blue"))
console.print("Loading in data from FreeMoCap SessionID: " + sessionID)
console.rule(style="color({})".format("blue"))

mediapipe_skel_fr_mar_dim = np.load(str(mediapipe_skel_data_path))
mediapipe_skel_fr_mar_dim = (
    mediapipe_skel_fr_mar_dim / 1000
)  # convert from mm to meters

(
    number_of_frames,
    number_of_tracked_points,
    number_of_dimensions,
) = mediapipe_skel_fr_mar_dim.shape
# load default mediapipe layout definitions

mediapipe_definitions_json_path = Path(
    r"C:/Users/jonma/Dropbox/GitKrakenRepos/freemocap/freemocap/dev_scratchpad/dev_alpha/default_json_definitions/freemocap_mediapipe_json.json"
)

with open(mediapipe_definitions_json_path) as json_file:
    mediapipe_definitions_dict = json.load(json_file)

tracked_point_names = mediapipe_definitions_dict["body"]["tracked_point_names"]

bpy.ops.object.empty_add(type="ARROWS", location=(0, 0, 0))

# plop empties at origin with names from mediapipe
empty_scale_factor = 0.05

for this_tracked_point_number in range(len(tracked_point_names)):
    bpy.ops.object.add(type="EMPTY")
    this_empty = bpy.context.active_object
    this_empty = bpy.context.active_object
    this_empty.name = tracked_point_names[this_tracked_point_number]
    this_empty.scale = (empty_scale_factor, empty_scale_factor, empty_scale_factor)


bpy.context.scene.unit_settings.length_unit = "METERS"
bpy.ops.object.mode_set(mode="OBJECT")


for frame_number in track(
    range(number_of_frames),
    description="[cyan]Loading tracked point trajectories as keyframed empties...",
    total=mediapipe_skel_fr_mar_dim.shape[0],
):
    bpy.context.scene.frame_set(frame_number)
    for this_empty_name in tracked_point_names:
        this_empty = bpy.data.objects[this_empty_name]
        this_empty.location = (
            mediapipe_skel_fr_mar_dim[
                frame_number, tracked_point_names.webcam_id(this_empty_name), 0
            ],
            mediapipe_skel_fr_mar_dim[
                frame_number, tracked_point_names.webcam_id(this_empty_name), 1
            ],
            mediapipe_skel_fr_mar_dim[
                frame_number, tracked_point_names.webcam_id(this_empty_name), 2
            ],
        )
        this_empty.keyframe_insert(data_path="location", frame=frame_number)


####################
### There are now EMPTIES in the scene!
####################
console.rule(style="color({})".format("magenta"))
console.print("[magenta] Generating Meta Rig \o/")
console.rule(style="color({})".format("magenta"))

bpy.ops.object.armature_human_metarig_add()
metarig = bpy.context.active_object

metarig_scale = 0.85
metarig.scale = (metarig_scale, metarig_scale, metarig_scale)


bpy.ops.pose.rigify_generate()
console.rule(style="color({})".format("magenta"))

console.rule(style="color({})".format("yellow"))
console.print("[yellow] Driving limb FK whosits with empty location data")
console.rule(style="color({})".format("yellow"))


## if not nan
## 1) copy_location - empty -> limb_FK & limb_IK(?)
## 2) set bpy.context.object.pose.bones["upper_arm_parent.L"].IK_FK = 1 (will later set this via reproj error or something)
## 3) add keyframes after each step with bpy.ops.anim.keyframe_insert_button(all=True)
## if nan
## 1) i think just skip? later we can interpolate IK/FK and whatnot
## on first frame
## set all IK stretch to 0
#
# for frame_number in track(range(number_of_frames), description="[yellow]Driving rig FK controllers with empty locations...", total=mediapipe_skel_fr_mar_dim.shape[0] ):
#    bpy.context.scene.frame_set(frame_number)
#    for this_empty_number, this_empty_name in enumerate(tracked_point_names):
#
#        this_point_connections_dict = mediapipe_definitions_dict["body"]['tracked_point_to_rigify_fk_rig_bone_connections'][this_empty_name]
#
#        this_empty_location = Vector(mediapipe_skel_fr_mar_dim[frame_number,this_empty_number,:])
#
#        if this_point_connections_dict:
#            for this_bone_name, this_bone_connection_dict in this_point_connections_dict.items():
#
#                bpy.ops.object.mode_set(mode='EDIT')
#                if this_bone_connection_dict["track_type"] == "bone_tail":
#                    if this_bone_name == "head" or this_bone_name == "toe":
#                        this_bone_obj = bpy.data.objects["rig"].data.bones[this_bone_name]
##                        this_bone_obj.tail = this_empty_location
#                        inspect(this_bone_obj, methods=True)
#                        break
#
#
#



from typing import NamedTuple

from src.cameras.multicam_manager import CVCameraManager
from src.core_processor.camera_calibration.charuco_board_detection.charuco_board_detector import CharucoBoardDetector


class ImageResponse(NamedTuple):
    image: bytes
    webcam_id: str


class BoardDetectService:
    def __init__(self, cam_manager: CVCameraManager = CVCameraManager()):
        self._cam_manager = cam_manager

    async def run(self):
        return CharucoBoardDetector(self._cam_manager).process()

    async def run_detection_on_cam_id(self, webcam_id: str, cb=None):
        await CharucoBoardDetector(self._cam_manager).process_by_cam_id(webcam_id, cb)



"""
### Data Handler class for wrapping and organizing OpenPose data into a more selfcontained and accessible format.

### Example code:

import fmc_trackdata_handler as df
import openpose_data_mappings as data_mapping

### Make empty data structure
data_handler = df.FmcTracDataHandler()

### Set mapping
data_handler.set_parent_mapping(data_mapping.parent_mapping)
data_handler.set_point_name_mapping(data_mapping.point_name_mapping)

### Load openpose npy data into variable
import numpy as np
datafile_path = r"G:\Shared drives\000_projects\cl005_northeasternMarkerlessMocap\incoming\datasets\sesh_21-05-31_111833\DataArrays\openPoseSkel_3d.npy"
data = np.load(datafile_path)

### Import data into data handler as an actor.
# This step can be repeated for multiple datasets, adding each as an individual actor
actor_name = "tester"
data_handler.import_actor_raw_data(actor_name, data)

### Saving data handler object to file
file_path = "/absolute/path/to/file/filename" # Suffix is being set to FILE_SUFFIX automatically
save_obj_to_file(file_path, data_handler, override=False)

### Loading data handler from file
file_path = "/absolute/path/to/file/filename.fmcData"
data_h = df.load_obj_from_file(file_path)

"""


import numpy as np
from pathlib import Path
import pickle

FILE_SUFFIX = ".fmcData"

class FmcTracDataHandler(object):
    """
    # data structure:
    {
    "header":{
        "version": str(),
        "tags": list(),
        "author": str(),
        "export_date": datetime(),
        "camera_count": int(),
        "license": str(),
        "calibration_obj": class,
        },
    "content":{
        str(actor_name): {
            "sample_count": int(),
            "tracking_points": {
                str(tracking_point_name): {"parents": list(), samples: list()}
            }
        }
    },
    }
    """

    def __init__(self):
        self.data = {"header":{}, "content":{}}

    def has_actor(self, actor_name):
        return actor_name in self.data["content"]

    def has_trackingpoint_name(self, actor_name, point_name):
        return self.has_actor(actor_name) and point_name in self.data["content"][actor_name]["tracking_points"]

    def set_point_name_mapping(self, point_name_mapping):
        self.point_name_mapping = point_name_mapping

    def set_parent_mapping(self, parent_mapping):
        self.parent_mapping = parent_mapping

    def init_actor(self, actor_name):
        if not self.has_actor(actor_name):
            self.data["content"][actor_name] = {}
        else:
            raise Exception("Actor already exist in dataset. Please remove it before re-adding")

    def init_tracking_point_name(self, actor_name, point_name):
        if not self.has_actor(actor_name):
            self.init_actor(actor_name)
        if not "tracking_points" in self.data["content"][actor_name]:
            self.data["content"][actor_name]["tracking_points"] = {}
        self.data["content"][actor_name]["tracking_points"][point_name] = {}

    def set_actor_data(self, actor_name, data_label, data):
        if not self.has_actor(actor_name):
            self.init_actor(actor_name)
        self.data["content"][actor_name][data_label] = data


    def set_tracking_point_samples(self, actor_name, point_name, samples):
        if not self.has_trackingpoint_name(actor_name, point_name):
            self.init_tracking_point_name(actor_name, point_name)
        self.data["content"][actor_name]["tracking_points"][point_name]["samples"] = samples

    def set_tracking_point_parents(self, actor_name, point_name, parents):
        if not self.has_trackingpoint_name(actor_name, point_name):
            self.init_tracking_point_name(actor_name, point_name)
        self.data["content"][actor_name]["tracking_points"][point_name]["parents"] = parents

    def set_actor_sample_count(self, actor_name, sample_count):
        if self.has_actor(actor_name):
            self.data["content"][actor_name]["sample_count"] = sample_count
        else:
            raise ValueError("Object does not contain an actor called: %s" % actor_name)

    def set_header(self,key, value):
        self.data["header"][key] = value

    def set_version(self, version):
        self.set_header("version", version)

    def set_tags(self, tag_list):
        self.set_header("tags", tag_list)

    def set_author(self, author):
        self.set_header("author", author)

    def set_date(self, datetime_obj):
        self.set_header("export_date", datetime_obj)

    def set_camera_count(self, camera_count):
        self.set_header("camera_count", camera_count)

    def set_license(self, license_string):
        self.set_header("license", license_string)

    def set_calibration_obj(self, calibration_obj):
        self.set_header("calibration_obj", calibration_obj)

    def import_actor_raw_data(self, actor_label, raw_data):
        mapped_data = self.map_point_names(raw_data)
        parent_mapping = self.get_parent_mapping()

        sample_count = False
        for point_name in mapped_data:
            parent_list = []
            if point_name in parent_mapping:
                parent_list.append(parent_mapping[point_name])

            #print("Trackingpoint: %s" % point_name)
            samples = mapped_data[point_name]
            self.set_tracking_point_samples(actor_label, point_name, samples)
            self.set_tracking_point_parents(actor_label, point_name, parent_list)

            if sample_count == False:
                sample_count = len(samples)
            else:
                if not sample_count == len(samples):
                    raise NotImplementedError("Tracking points have different sample count. This is not supported yet")
                else:
                    sample_count = len(samples)
        self.set_actor_sample_count(actor_label, sample_count)


    def get_data(self):
        return self.data

    def get_actor_data(self, actor_label):
        if self.has_actor(actor_name):
            return self.data["content"][actor_label]
        else:
            raise ValueError("Object has no actor named %s" % actor_label)

    def list_actors(self):
        return list(self.data["content"].keys())

    def get_actor_tracking_points(self, actor_name):
        return self.data["content"][actor_name]["tracking_points"]

    def get_point_parents(self, actor_name, point_name):
        return self.data["content"][actor_name]["tracking_points"][point_name]["parents"]

    def get_actor_sample_count(self, actor_name):
        if self.has_actor(actor_name):
            return self.data["content"][actor_name]["sample_count"]
        else:
            raise ValueError("Object does not contain an actor called: %s" % actor_name)





    ### OpenPose mapping
    ################################################################################
    def map_point_names(self, data):
        """
        input data of format:
            [sample_idx][point_index] = (x,y,z)

        output format
            {point_name:
                [(x,y,z), (x,y,z),...]} # list of samples
        """
        point_names = self.get_point_name_mapping()

        # initiating data container for each point in point_names, added generated names for indices beyond length of point_names
        first_sample = data[0]
        if len(first_sample) > len(point_names): # Check if there is an equal amount of names points and indices in the data
            surplus_sample_points = len(first_sample) - len(point_names)
            for i in range(surplus_sample_points):
                point_names.append("generatedName%s" % str(i).zfill(3))

        # reordering data to be ["point_name"][sample_i][x,y,z]
        data_reform = {}
        for i, point_name in enumerate(point_names):
            data_reform[point_name] = data[:,i,:]

        return data_reform



    def get_parent_mapping(self):
        """
        Returns a dictionary that maps point names to their parent point
        """
        if not self.parent_mapping == False:
            return self.parent_mapping
        else:
            raise ValueError("No parent mapping is set. Use set_parent_mapping() to define an appropriate map")

    def get_point_name_mapping(self):
        """
        Returns a list of names corresponding to the data positions
        """
        if not self.point_name_mapping == False:
            return self.point_name_mapping
        else:
            raise ValueError("No point name mapping is set. Use set_point_name_mapping() to define an appropriate map")





def load_obj_from_file(file_path):
    # Loads file saved through this class
    f = Path(file_path)

    if f.is_file():
        if f.suffix.lower() == FILE_SUFFIX.lower():
            print(f"Reading file from: {f.absolute()}")
            with open(file_path, "rb") as infile:
                data = pickle.load(infile)
            if not isinstance(data, FmcTracDataHandler):
                raise ValueError("Loaded data is not an instance of FmcTracDataHandler")
        else:
            raise ValueError("File has wrong extension. Should be '%s'" % FILE_SUFFIX)
    else:
        raise ValueError("File does not exist")

    return data




def save_obj_to_file(file_path, fmxDataObj, override=False):
    f = Path(file_path)
    f= f.with_suffix(FILE_SUFFIX)
    if f.exists() and override == False:
        raise NotImplementedError(f"File already exists and override set to False, Skipping\nFilepath: {f.absolute()}")
    with open(f, 'wb') as handle:
        pickle.dump(fmxDataObj, handle, protocol=2)
        print(f"Saved file to: {f.absolute()}")

f = 2
# import fmc_trackdata_handler as df
# import openpose_data_mappings as data_mapping

# ## Make empty data structure
# data_handler = df.FmcTracDataHandler()

# ## Set mapping
# data_handler.set_parent_mapping(data_mapping.parent_mapping)
# data_handler.set_point_name_mapping(data_mapping.point_name_mapping)

# ## Load openpose npy data into variable
# import numpy as np
# datafile_path = r"C:\Users\Rontc\Documents\GitHub\freemocap\FreeMocap_Data\sesh_21-07-29_104227\DataArrays\openPoseSkel_3d.npy"
# data = np.load(datafile_path)

# ## Import data into data handler as an actor.
# #This step can be repeated for multiple datasets, adding each as an individual actor
# actor_name = "tester"
# data_handler.import_actor_raw_data(actor_name, data)

# ## Saving data handler object to file
# file_path = r"C:\Users\Rontc\Documents\GitHub\freemocap\FreeMocap_Data\sesh_21-07-29_104227\DataArrays\openPoseSkel_3d" # Suffix is being set to FILE_SUFFIX automatically
# df.save_obj_to_file(file_path, data_handler, override=False)

# ## Loading data handler from file
# file_path = r"C:\Users\Rontc\Documents\GitHub\freemocap\FreeMocap_Data\sesh_21-07-29_104227\DataArrays\openPoseSkel_3d.fmcData"
# data_h = df.load_obj_from_file(file_path)

# f = 2




from typing import List

from pydantic import BaseModel


class RawCamera(BaseModel):
    webcam_id: str


class FoundCamerasResponse(BaseModel):
    number_of_cameras_found: int
    cameras_found_list: List[RawCamera]
    cv2_backend: int



from pathlib import Path
import copy 

import numpy as np
from rich.progress import Progress
import cv2
import mediapipe as mp

# from numba import jit

mp_drawing = mp.solutions.drawing_utils
mp_drawing_styles = mp.solutions.drawing_styles
mp_holistic = mp.solutions.holistic


def run_mediapipe(synched_video_path_list, dummyRun=False):
    """ 
    Run MediaPipe on synced videos, and save body tracking data to be parsed 
    """  
    num_vids = len(synched_video_path_list)
    
    with mp_holistic.Holistic(
                            static_image_mode = False, #use 'static image mode' to avoid system getting 'stuck' on ghost skeletons?
                            model_complexity = 2, #in this house, we turn the Speed/Accuracy dial all the way towards accuracy \o/)
                            ) as holistic:

        mediapipe_results_from_all_videos_list = []  # Create an empty list that holds each cameras data
        each_video_resolution_height = [] # mediapipe gives results in normalized screen coordinate, so we need this to convert to pixels
        each_video_resolution_width = []
        
        for this_vid_path in ( synched_video_path_list):  # Run MediaPipe 'Holistic' (body, hands, face) tracker on each video in the list of synched videos
            mediaPipe_results_from_this_video_list = []  # Create an empty list for mediapipes data
            if not dummyRun:
                cap = cv2.VideoCapture(str(this_vid_path))

                frame_num = -1
                num_frames = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
                success, image_raw = cap.read()  # load first image from video

                with Progress() as progress:
                    progressBar = progress.add_task("[magenta]MediaPiping: {}".format(this_vid_path.name),total=num_frames,)  # make progressbar

                    while success:

                        if frame_num % 5 == 0:
                            progress.update( progressBar, advance=5 )  # increment progress bar everyh 5th frame

                        frame_num += 1

                        image_height, image_width, _ = image_raw.shape
                        image = cv2.cvtColor(image_raw, cv2.COLOR_BGR2RGB), #Convert the BGR image to RGB before processing
                        image = image[0] #unpack the tuple into a straight up image
                        results = holistic.process( image)  # NOTE: THIS IS WHERE THE MAGIC HAPENS 

                        mediaPipe_results_from_this_video_list.append( results )  # Append data to mediapipe data list

                        # load the next image (will break `while-loop` on the last frame)
                        success, image_raw = cap.read()  # load next image from video

            each_video_resolution_height.append(image_height)
            each_video_resolution_width.append(image_width)

            mediapipe_results_from_all_videos_list.append( mediaPipe_results_from_this_video_list)  # Append that cameras data for every frame to the camera datalist

    ##  
    ##  
    ##  #Parse MediaPipe into FMC Format
    ##  
    ##  
    
    #standard MediaPipe output
    numBodyPoints = 33
    numFacePoints = 468 #ton of face points because it's a mesh (rather than individually tracked points)
    numHandPoints = 21 #remember, your average person has almost exactly 2 hands

    numTrackedPoints = (numBodyPoints + numHandPoints * 2 + numFacePoints )  # Get total points

    # Create  array of nans the size of number of cams, frame, points, XYC
    mediaPipeData_nCams_nFrames_nImgPts_XYC = np.empty((int(num_vids), int(num_frames), int(numTrackedPoints), 3))  # create empty array
    mediaPipeData_nCams_nFrames_nImgPts_XYC[:] = np.NaN  # Fill it with NaNs!

    list_of_frames_with_full_data = []
    
    for vid_num in range(num_vids):  # Loop through each camera
        this_vid_mediapipe_results_per_frame = mediapipe_results_from_all_videos_list[vid_num]
        for frame_num in range(num_frames):  # Loop through each frame
            
            this_frame_mediapipe_results  = this_vid_mediapipe_results_per_frame[frame_num]
            
            thisFrame_X_body = np.empty(numBodyPoints)
            thisFrame_X_body[:] = np.nan
            thisFrame_Y_body = thisFrame_X_body.copy()
            thisFrame_C_body = thisFrame_X_body.copy() #'C' is for 'condfidence' (mediapipe calls it 'visibility')

            thisFrame_X_face = np.empty(numFacePoints)
            thisFrame_X_face[:] = np.nan
            thisFrame_Y_face = thisFrame_X_face.copy()
            thisFrame_C_face = thisFrame_X_face.copy()

            thisFrame_hands = np.empty(numHandPoints)
            thisFrame_hands[:]= np.nan

            thisFrame_X_lefthand = thisFrame_hands.copy()
            thisFrame_Y_lefthand = thisFrame_hands.copy()
            thisFrame_C_lefthand = thisFrame_hands.copy()

            thisFrame_X_righthand = thisFrame_hands.copy()
            thisFrame_Y_righthand = thisFrame_hands.copy()
            thisFrame_C_righthand = thisFrame_hands.copy()

            fullFrame = True #keep track of which frames have 'full' data (all body, all hands, all face data present)
            try:
                # pull out ThisFrame's mediapipe data (`mpData.pose_landmarks.landmark` returns something iterable ¯\_(ツ)_/¯)
                thisFrame_poseDataLandMarks = this_frame_mediapipe_results.pose_landmarks.landmark  # body ('pose') data
                # stuff body data into pre-allocated nan array
                thisFrame_X_body[:numBodyPoints] = [pp.x for pp in thisFrame_poseDataLandMarks]  # PoseX data - Normalized screen coords (in range [0, 1]) - need multiply by image resultion for pixels
                thisFrame_Y_body[:numBodyPoints] = [pp.y for pp in thisFrame_poseDataLandMarks]  # PoseY data
                thisFrame_C_body[:numBodyPoints] = [pp.visibility for pp in thisFrame_poseDataLandMarks]  #'visibility' is MediaPose's 'confidence' measure in range [0,1]
            except:
                fullFrame = False

            # Right hand data
            try:
                thisFrame_rHandDataLandMarks = this_frame_mediapipe_results.right_hand_landmarks.landmark  # right hand data
                thisFrame_X_righthand[:numHandPoints] = [pp.x for pp in thisFrame_rHandDataLandMarks]  # PoseX data - Normalized screen coords (in range [0, 1]) - need multiply by image resultion for pixels
                thisFrame_Y_righthand[:numHandPoints] = [pp.y for pp in thisFrame_rHandDataLandMarks]  # PoseY data
                thisFrame_C_righthand[:numHandPoints] = [pp.visibility for pp in thisFrame_rHandDataLandMarks]  #'visibility' is MediaPose's 'confidence' measure in range [0,1]
            except:
                fullFrame = False

            # Left hand data
            try:
                thisFrame_lHandDataLandMarks = this_frame_mediapipe_results.left_hand_landmarks.landmark  # left hand data
                thisFrame_X_lefthand[:numHandPoints ] = [pp.x for pp in thisFrame_lHandDataLandMarks]  # PoseX data - Normalized screen coords (in range [0, 1]) - need multiply by image resultion for pixels
                thisFrame_Y_lefthand[:numHandPoints] = [pp.y for pp in thisFrame_lHandDataLandMarks]  # PoseY data
                thisFrame_C_lefthand[:numHandPoints] = [pp.visibility for pp in thisFrame_lHandDataLandMarks]  #'visibility' is MediaPose's 'confidence' measure in range [0,1]
            except:
                fullFrame = False

            # FaceMeshData
            try:
                thisFrame_faceDataLandMarks = this_frame_mediapipe_results.face_landmarks.landmark  # face (mesh) data
                thisFrame_X_face[:numFacePoints] = [pp.x for pp in thisFrame_faceDataLandMarks]  # PoseX data - Normalized screen coords (in range [0, 1]) - need multiply by image resultion for pixels
                thisFrame_Y_face[:numFacePoints] = [pp.y for pp in thisFrame_faceDataLandMarks]  # PoseY data
                # NOTE - There's also Z data in here, but we're gonna ignore it
                thisFrame_C_face[:numFacePoints] = [ pp.visibility for pp in thisFrame_faceDataLandMarks ]  #'visibility' is MediaPose's 'confidence' measure in range [0,1]
            except:
                fullFrame = False

            if fullFrame:
                list_of_frames_with_full_data.append(frame_num)

            thisFrame_X = np.concatenate((thisFrame_X_body,thisFrame_X_righthand,thisFrame_X_lefthand,thisFrame_X_face))
            thisFrame_Y = np.concatenate((thisFrame_Y_body,thisFrame_Y_righthand,thisFrame_Y_lefthand,thisFrame_Y_face))
            thisFrame_C = np.concatenate((thisFrame_C_body,thisFrame_C_righthand,thisFrame_C_lefthand,thisFrame_C_face))
            
            # stuff this frame's data into pre-allocated mediaPipeData_.... array
            mediaPipeData_nCams_nFrames_nImgPts_XYC[vid_num, frame_num, :, 0] = thisFrame_X *each_video_resolution_width[vid_num]# convert from normalized screen coordinates to pixel coordinates
            mediaPipeData_nCams_nFrames_nImgPts_XYC[vid_num, frame_num, :, 1] = thisFrame_Y *each_video_resolution_height[vid_num]
            mediaPipeData_nCams_nFrames_nImgPts_XYC[vid_num, frame_num, :, 2] = thisFrame_C


    mediaPipeData_nCams_nFrames_nImgPts_XYC[:, :, 34:, 2] = 1 #hand and face data don't get 'visibility'/'confidence' scores, so we'll just set them all to 1

    return mediaPipeData_nCams_nFrames_nImgPts_XYC




if __name__ == "__main__":
    synched_vid_folder_path = Path(r"C:\Users\jonma\Dropbox\FreeMoCapProject\FreeMocap_Data\sesh_2021-12-06_08_52_10_mc\SyncedVideos")
    synched_video_path_list = [vid_path for vid_path in synched_vid_folder_path.glob("*.mp4")]
    
    mediaPipeData_nCams_nFrames_nImgPts_XYC = run_mediapipe(synched_video_path_list=synched_video_path_list)
    mediapipe2d_filename = synched_vid_folder_path.parent/'mediapipe_2d_points.npy'
    np.save(mediapipe2d_filename, mediaPipeData_nCams_nFrames_nImgPts_XYC)
    
    


import numpy as np


class TimestampLogger:
    pass
    # def __init__(self, _camera_id_list: list[str], camera0_id: str = '0'):
    #     self._camera0_id = camera0_id
    #     self._number_of_cameras = len(_camera_id_list)
    #     self._camera_timestamp_dict = {}
    #
    #     for this_cam_id in _camera_id_list:
    #         self._camera_timestamp_dict[this_cam_id] = np.ndarray(0)
    #
    # def update(self, this_webcam_id, new_timestamp):
    #     self._camera_timestamp_dict[this_webcam_id] = np.append(self._camera_timestamp_dict[this_webcam_id],
    #                                                             new_timestamp)
    #
    #     timestamp_array_lengths = [this_timestamp_array.shape[0] for this_timestamp_array in self._camera_timestamp_dict.values()]
    #
    #     if np.sum(np.diff(timestamp_array_lengths)) == 0:  # if timestamp arrays are all the same length
    #         self._update_timestamp_differences_from_camera()
    #



#%%
import logging
from typing import List
import threading
import queue
import time
import datetime

from pathlib import Path

from pydantic import BaseModel
import cv2
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd

from opencv_camera import OpenCVCamera

from rich import print, inspect




logger = logging.getLogger(__name__)
logger.level = logging.INFO

MAX_NUMBER_OF_CAMERAS = 20

freemocap_data_path = Path('C:/Users/jonma/Dropbox/FreeMoCapProject/FreeMocap_Data')
sessionID = datetime.datetime.now().strftime("sesh_%Y-%m-%d_%H_%M_%S_mch")
session_path = freemocap_data_path / sessionID
session_path.mkdir(parents=True, exist_ok=True)

def stuff_incoming_frames_into_a_queue(camera: OpenCVCamera, thread_queue: queue.Queue):

    print('stuff_incoming_frames_into_a_queue starting for {}'.format(camera.name))

    while thread_exit_event.is_set() is False:
        success, image, timestamp_ns = camera.get_next_frame()
        logger.info('{} grabbed a frame at timestamp {}'.format(camera.name, timestamp_ns/1e9))
        camera.opencv_video_writer_object.write(image)
        if not success:
            logger.error('{} failed to grab a frame at timestamp {}'.format(camera.name, timestamp_ns/1e9))
        thread_queue.put((success, camera.port_number, camera.name, image, timestamp_ns))




def start_camera_thread(camera: OpenCVCamera, incoming_frames_queue: queue.Queue):
    print('run_camera_in_thread starting for {}'.format(camera.name))
    cam_vid_filename = camera.name +'_synchronized.mp4'
    camera_save_path = session_path / cam_vid_filename
    camera.initialize_video_writer(str(camera_save_path))
    camera_thread = threading.Thread(
                                target=stuff_incoming_frames_into_a_queue,
                                args=(camera, incoming_frames_queue),
                                name=camera.name+'-thread')
    camera_thread.start()
    return camera_thread

def display_image_on_its_cv2_named_window(cam_window_name: str, image: np.ndarray, ):
    cv2.imshow(cam_window_name, image)
    key = cv2.waitKey(1)

    if key == 27: #esc key kills all streams, I think
        thread_exit_event.set()

    if thread_exit_event.is_set():
        cv2.destroyAllWindows()

def create_video_viewing_window(camera_name:str = 'Cam', window_location = (0,0)):
    window_name = camera_name + " - Press ESC to exit"
    cv2.namedWindow(window_name)
    cv2.moveWindow(window_name, window_location[0], window_location[1])
    return window_name




if __name__ == '__main__':
    #%% connect to all available cameras

    camera_found_bool = True
    camera_list = []
    this_port_number = -1
    while camera_found_bool and len(camera_list) < 5:
        this_port_number+=1
        camera =  OpenCVCamera(port_number=this_port_number)
        camera_found_bool = camera.connect()
        if camera_found_bool:
            camera_list.append(camera)

    number_of_cameras = len(camera_list)

    
    global thread_exit_event
    thread_exit_event = threading.Event()
    loop_count = 0
    camera_thread_list = [None]*len(camera_list)
    incoming_frames_queue = queue.Queue()

    #%% set some things up before we start streaming
    # setting up to grab incoming timestamps etc
    dict_of_timestamp_lists = {}
    dict_of_fps_mean_lists = {}
    dict_of_fps_std_lists = {}
    for this_camera in camera_list:
        dict_of_timestamp_lists[this_camera.name] = []
        dict_of_fps_mean_lists[this_camera.name] = []
        dict_of_fps_std_lists[this_camera.name] = []

    # #create cv2 named windows where we'll display incoming frames
    cam_window_name_dict = {}
    display_image_scale_factor = 2
    scaled_image_width = int(camera_list[0].resolution_width/display_image_scale_factor)
    scaled_image_height = int(camera_list[0].resolution_height/display_image_scale_factor)
    window_buffer = 50
    for this_camera in camera_list:
        #tile windows across monitor
        window_location_x = this_camera.port_number*scaled_image_width
        window_location_y=0
        if window_location_x+scaled_image_width > 1920:
            window_location_x -= 1920
            window_location_y = scaled_image_height+window_buffer
        elif window_location_x+scaled_image_width > 1920*2:
            window_location_x -= 1920*2
            window_location_y = scaled_image_height*2
        print('Creating window for {} at ({},{})'.format(this_camera.name, str(window_location_x), str(window_location_y)))
        cam_window_name_dict[this_camera.name] = create_video_viewing_window(this_camera.name, (window_location_x, window_location_y))

    # multi_cam_widow_name = create_video_viewing_window('MultiCamera')

    # #were going to scale down each incoming camera image and tile it, so we only have to display one big thing (which is faster than displaying many small things)
    # multi_image_number_of_tiles_per_side = int(np.ceil(np.sqrt(number_of_cameras)))
    # num_extra_spots_in_multi_image = multi_image_number_of_tiles_per_side - number_of_cameras
    # display_image_scale_factor = 2
    # scaled_image_width = int(camera_list[0].resolution_width/display_image_scale_factor)
    # scaled_image_height = int(camera_list[0].resolution_height/display_image_scale_factor)

    # scaled_image_black_frame = np.zeros((scaled_image_height, scaled_image_width, 3))
    # multi_image_holder_list_of_lists = []

    # for iter in range(multi_image_number_of_tiles_per_side):
    #     blank_row = [scaled_image_black_frame.copy() for each in range(multi_image_number_of_tiles_per_side)]
    #     multi_image_holder_list_of_lists.append(blank_row)

    # each_cameras_multi_image_address = []
    # for this_cam_num in range(number_of_cameras):
    #     this_cam_row = int(np.floor(this_cam_num/multi_image_number_of_tiles_per_side))
    #     this_cam_col = this_cam_num - this_cam_row*multi_image_number_of_tiles_per_side
    #     each_cameras_multi_image_address.append((this_cam_row, this_cam_col))



    #%%
    ###########################################################################################
    ###########################################################################################
    ##
    ##  Start the camera threads (THIS IS WHERE THE MAGIC HAPPENS)
    ##
    ###########################################################################################
    ###########################################################################################

    for this_camera in camera_list:
        camera_thread_list[this_camera.port_number] = start_camera_thread(this_camera, incoming_frames_queue)

    #
    # start the MainThread loop
    approx_start_time = time.time()
    end_after_seconds = 20 #or on pressing ESC
    qsize = []
    while not thread_exit_event.is_set():
        time_elapsed = time.time() - approx_start_time
        this_qsize = incoming_frames_queue.qsize()
        qsize.append(this_qsize)
        print('Time remaining: {:3.2f} - Queueue size: {:3.0f} '.format(end_after_seconds - time_elapsed, this_qsize))
        if time_elapsed > end_after_seconds:
            thread_exit_event.set()
            cv2.destroyAllWindows()

        if not incoming_frames_queue.empty():
            this_frame_tuple = incoming_frames_queue.get() #<- grab the most recent frame_tuple and do stuff with it
            this_success_bool, this_camera_number, this_camera_name, this_image, this_timestamp_ns = this_frame_tuple
            this_timestamp_sec = this_timestamp_ns/1e9
            dict_of_timestamp_lists[this_camera_name].append(this_timestamp_sec)

        elif qsize[-1] < 10: #only display images if queue isn't over filled
            # this_cam_address = each_cameras_multi_image_address[this_camera_number]
            # print(this_cam_address)
            # resized_image = cv2.resize(this_image, (scaled_image_width, scaled_image_height))
            # multi_image_holder_list_of_lists[this_cam_address[0]][this_cam_address[1]] = resized_image
            
            # each_row_of_multi_im = []
            # for this_row in range(multi_image_number_of_tiles_per_side):
            #     each_row_of_multi_im.append(np.hstack(multi_image_holder_list_of_lists[this_row]))
            # this_multi_im = np.vstack(each_row_of_multi_im)
            # this_multi_im = cv2.normalize(src=this_multi_im, dst=None, alpha=0, beta=255, norm_type=cv2.NORM_MINMAX, dtype=cv2.CV_8U)
            # display_image_on_its_cv2_named_window(multi_cam_widow_name, this_multi_im)
            image_height, image_width, image_channels = this_image.shape
            scale_height = int(image_height/display_image_scale_factor)
            scale_width = int(image_width/display_image_scale_factor)
            display_image_on_its_cv2_named_window(cam_window_name_dict[this_camera_name], cv2.resize(this_image, (scale_width, scale_height)))

    #%%

    list_lengths = [len(this_thing) for this_thing in dict_of_timestamp_lists ]
    for this_camera in camera_list:
        this_cam_timestamps = np.array(dict_of_timestamp_lists[this_camera.name])

    number_of_frames = np.min([len(dict_of_timestamp_lists[this_key]) for this_key in dict_of_timestamp_lists ]) #truncate all timestamp lists to match frame count of shortest
    out_timestamp_dict = {}

    for this_cam_num, this_camera in enumerate(camera_list):
        out_timestamp_dict[this_camera.name] = np.array(dict_of_timestamp_lists[this_camera.name][:number_of_frames])
    out_dataframe = pd.DataFrame(out_timestamp_dict)
    out_dataframe.to_csv('~/timestamps.csv')

    #%% Plot up them bad bois
    # where 'bad boi'  is defined as a timestamp from each camera
    # plt.ion() #<- allows for fiddling, but need to set breakpoint and call plt.show() to see the plots
    figure = plt.figure(3214, figsize=(18,5))
    ax_timestamps = figure.add_subplot(141)
    ax_timestamps.set_xlabel('frame number')
    ax_timestamps.set_ylabel('timestamp (unix epoch, sec)')

    ax_frame_duration = figure.add_subplot(142)
    ax_frame_duration.set_xlabel('frame number')
    ax_frame_duration.set_ylabel('framerate (1/np.diff(timestamp, frames/sec)')
    ax_frame_duration.set_ylim(0, .1)

    ax_histogram = figure.add_subplot(143)
    ax_histogram.set_xlabel('bins (fps)')
    ax_histogram.set_ylabel('histogram count(out of {} frames'.format(number_of_frames))

    ax_qsize = figure.add_subplot(144)
    ax_qsize.set_xlabel('frame/loop count')
    ax_qsize.set_ylabel('incoming frames queueue size')
    ax_qsize.plot(qsize, '-',linewidth=.5, alpha=.5)
    ax_qsize.plot(qsize, 'k.')

    for this_camera in camera_list:
        this_cam_timestamps = out_timestamp_dict[this_camera.name]
        this_cam_frame_duration = np.diff(out_timestamp_dict[this_camera.name])
        this_cam_framerate = 1/this_cam_frame_duration

        ax_timestamps.plot(this_cam_timestamps, 'o',markersize=1, alpha=.5, label=this_camera.name)
        ax_frame_duration.plot(this_cam_frame_duration, '-o', markersize=1, alpha=.5, label=this_camera.name)
        ax_histogram.hist(this_cam_framerate, bins=np.arange(0,50,1), alpha=.5,label=this_camera.name)

    #reference lines (i.e. 'ideal' numbers)
    ideal_framerate = 30
    approx_start_time = this_cam_timestamps[0] #just for the last camera, but it's fine for this visualization
    ideal_end_time = approx_start_time + number_of_frames/ideal_framerate
    ideal_timestamps = np.linspace(approx_start_time, ideal_end_time, number_of_frames)

    ideal_line_label = 'ideal for {}fps'.format(ideal_framerate)
    ax_frame_duration.plot([ 0, number_of_frames], [1/ideal_framerate, 1/ideal_framerate],'--', label=ideal_line_label, color='black')
    ax_histogram.vlines(ideal_framerate, 0, number_of_frames, linestyles='dashed', label=ideal_line_label)
    ax_timestamps.plot(ideal_timestamps, '--', color='black', label=ideal_line_label)
    ax_frame_duration.legend()

    plt.show()
    plt.pause(0.1)#put a breakpoint here to view/interact with figure if `plt.ion()` is active




from pathlib import Path

from pydantic import BaseModel

def _get_home_dir():
    return str(Path.home())

class WebcamConfig(BaseModel):
    webcam_id: int = 0
    exposure: int = -6
    resolution_width: int = 800
    resolution_height: int = 600
    save_video: bool = True
    # fourcc: str = "MP4V"
    fourcc: str = "MJPG"
    base_save_video_dir = _get_home_dir()


import logging
import traceback

import cv2
from fastapi import APIRouter, WebSocket

from src.api.services.board_detect_service import BoardDetectService
from src.api.services.mediapipe_detect_service import MediapipeSkeletonDetectionService
from src.cameras.multicam_manager.cv_camera_manager import OpenCVCameraManager

logger = logging.getLogger(__name__)

cam_ws_router = APIRouter()


@cam_ws_router.get("/begin_board_detection")
async def begin_board_detection():
    service = BoardDetectService()
    await service.run()


@cam_ws_router.websocket("/ws/charuco_board_detection/{webcam_id}")
async def board_detection_as_ws(web_socket: WebSocket, webcam_id: str):
    await web_socket.accept()

    async def websocket_send(input_image):
        success, frame = cv2.imencode(".png", input_image)
        if success:
            await web_socket.send_bytes(frame.tobytes())

    try:
        await BoardDetectService().run_detection_on_cam_id(
            webcam_id=webcam_id, cb=websocket_send
        )
    except:
        logger.error("Websocket ended")
        traceback.print_exc()
        return


@cam_ws_router.websocket("/ws/skeleton_detection/{webcam_id}")
async def skeleton_detection_as_ws(
    web_socket: WebSocket, webcam_id: str, model_complexity: int
):
    await web_socket.accept()

    async def websocket_send(input_image):
        success, frame = cv2.imencode(".png", input_image)
        if success:
            await web_socket.send_bytes(frame.tobytes())

    try:
        await MediapipeSkeletonDetectionService(OpenCVCameraManager()).run_as_loop(
            webcam_id=webcam_id, cb=websocket_send, model_complexity=model_complexity
        )
    except:
        logger.error("Websocket ended")
        traceback.print_exc()
        return


@cam_ws_router.get("/begin_mediapipe_skeleton_detection")
async def begin_mediapipe_skeleton_detection(model_complexity: int):
    """
    model_complexity can be 1 (faster, less accurate) or 2 (slower, more accurate)
    """
    service = MediapipeSkeletonDetectionService(OpenCVCameraManager())
    service.run(model_complexity)



from freemocap import fmc_anipose
import numpy as np
from rich.console import Console
console=Console()

# from numba import jit

# @jit(nopython=True)
def reconstruct3D(session, data_nCams_nFrames_nImgPts_XYC, confidenceThreshold=0.3):
    """
    Take a specifically formatted data array, and based on the camera calibration yaml, reconstruct a 3D image
    """

    if (
        session.cgroup is None
    ):  # load the calibration settings into the session class if it hasn't been already
        calibrationFile = "{}_calibration.toml".format(session.sessionID)
        session.cameraConfigFilePath = session.sessionPath / calibrationFile
        session.cgroup = fmc_anipose.CameraGroup.load(session.cameraConfigFilePath)

    nCams, nFrames, nImgPts, nDims = data_nCams_nFrames_nImgPts_XYC.shape

    if nDims == 3:
        dataOG = data_nCams_nFrames_nImgPts_XYC.copy()

        for camNum in range(nCams):
                          
            thisCamX = data_nCams_nFrames_nImgPts_XYC[camNum, :, :,0 ]
            thisCamY = data_nCams_nFrames_nImgPts_XYC[camNum, :, :,1 ]
            thisCamConf = data_nCams_nFrames_nImgPts_XYC[camNum, :, :, 2]

            thisCamX[thisCamConf < confidenceThreshold] = np.nan
            thisCamY[thisCamConf < confidenceThreshold] = np.nan

            if session.debug:
                import matplotlib.pyplot as plt
                fig = plt.figure(8000+camNum)
                fig.suptitle("3d reconstruction Confidence thresholding - cam{}".format(camNum))               
                axOG = fig.add_subplot(1,2,1)
                axOG.imshow(dataOG[0,:,:,0])
                axOG.set_title("Original Data")
                axTh = fig.add_subplot(1,2,2)
                axTh.imshow(thisCamX)
                axTh.set_title("Thresholded Data (there should be NEW and EXCITING gaps :O")
                

                


    if nDims == 2:
        data_nCams_nFrames_nImgPts_XY = data_nCams_nFrames_nImgPts_XYC
    elif nDims == 3:
        data_nCams_nFrames_nImgPts_XY = np.squeeze(data_nCams_nFrames_nImgPts_XYC[:, :, :, 0:2])

    dataFlat_nCams_nTotalPoints_XY = data_nCams_nFrames_nImgPts_XY.reshape(nCams, -1, 2)  # reshape data to collapse across 'frames' so it becomes [numCams, numFrames*numPoints, XY]

    console.print('Reconstructing 3d points...')
    data3d_flat = session.cgroup.triangulate(dataFlat_nCams_nTotalPoints_XY, progress=True)

    dataReprojerr_flat = session.cgroup.reprojection_error( data3d_flat, dataFlat_nCams_nTotalPoints_XY, mean=True)

    ##return:
    data_fr_mar_xyz = data3d_flat.reshape(nFrames, nImgPts, 3)
    dataReprojErr = dataReprojerr_flat.reshape(nFrames, nImgPts)

    return data_fr_mar_xyz, dataReprojErr



##   
##   
##   ███████ ███    ███  ██████     ███    ███ ██    ██ ██      ████████ ██        ██████  █████  ███    ███ ███████ ██████   █████  
##   ██      ████  ████ ██          ████  ████ ██    ██ ██         ██    ██       ██      ██   ██ ████  ████ ██      ██   ██ ██   ██ 
##   █████   ██ ████ ██ ██          ██ ████ ██ ██    ██ ██         ██    ██ █████ ██      ███████ ██ ████ ██ █████   ██████  ███████ 
##   ██      ██  ██  ██ ██          ██  ██  ██ ██    ██ ██         ██    ██       ██      ██   ██ ██  ██  ██ ██      ██   ██ ██   ██ 
##   ██      ██      ██  ██████     ██      ██  ██████  ███████    ██    ██        ██████ ██   ██ ██      ██ ███████ ██   ██ ██   ██ 
##                                                                                                                                   
##                                                                                                                                   
#Font - ANSI Regular - https://patorjk.com/software/taag/#p=display&f=ANSI%20Regular&t=Play%20Skeleton%20Animation


from os import mkdir
from FMC_Camera import FMC_Camera

import concurrent.futures
import logging
import queue
import threading
import pathos.multiprocessing as pathos_mp
from pathos.helpers import mp as pathos_mp_helper
import time
import platform
import datetime

import cv2
import matplotlib.pyplot as plt
import numpy as np
import h5py
from pathlib import Path

from rich import print
from rich import inspect
from rich.progress import track
from rich.traceback import install
install(show_locals=True)
from rich.console import Console
rich_console = Console()
from rich.progress import track


class FMC_MultiCameraRecorder:
    ##  
    ##  
    ##                  ██ ███    ██ ██ ████████                 
    ##                  ██ ████   ██ ██    ██                    
    ##                  ██ ██ ██  ██ ██    ██                    
    ##                  ██ ██  ██ ██ ██    ██                    
    ##  ███████ ███████ ██ ██   ████ ██    ██    ███████ ███████ 
    ##                                                           
    ##  
    ##  
    def __init__(
                self,
                rec_name = None,
                save_path = None,
                num_cams = 0,
                cams_to_use_list = None,
                show_multi_cam_stream = True,
                save_to_mp4 = True,
                save_to_h5 = False,
                save_log_file = False
                ):
        
        self._init_start_time = time.time_ns()#the precision is aspirational, lol

        if rec_name is None:
            self._rec_name = datetime.datetime.now().strftime("FMC_MultiCamRecording_%Y-%b-%d_%H_%M_%S")
        else:
            self._rec_name = rec_name


        if save_path is None:
            self._save_path = Path('data/' + self._rec_name)
        else:
            self._save_path = Path(save_path)
        
        if save_path or save_to_mp4 or save_to_h5 or save_log_file:
            self._save_path.mkdir(parents=True)
            self._path_to_log_file = Path(str(self._save_path / self._rec_name) + '_log.txt')

        self._save_to_mp4 = save_to_mp4
        if self._save_to_mp4:            
            self._output_video_object = None #build this after we get the first multi_frame_image

        # log_file_obj = open(self._path_to_log_file,'w+')
        # rich_console = Console(file=log_file_obj)
        rich_console = Console()

        self._num_cams = num_cams
        self._cams_to_use_list = cams_to_use_list
        
        
        
        
        rich_console.rule('Starting FreeMoCap MultiCam Recorder!')


        

    ###   
    ###   
    ###   ██████  ██████   ██████  ██████  ███████ ██████  ████████ ██ ███████ ███████ 
    ###   ██   ██ ██   ██ ██    ██ ██   ██ ██      ██   ██    ██    ██ ██      ██      
    ###   ██████  ██████  ██    ██ ██████  █████   ██████     ██    ██ █████   ███████ 
    ###   ██      ██   ██ ██    ██ ██      ██      ██   ██    ██    ██ ██           ██ 
    ###   ██      ██   ██  ██████  ██      ███████ ██   ██    ██    ██ ███████ ███████ 
    ###                                                                                
    ###                                                                                

    @property
    def num_cams(self):
        """How many cameras to use"""        
        return self._num_cams

    @num_cams.setter
    def num_cams(self, num_cams_input):
        """Define how many cameras to use"""
        self._num_cams = num_cams_input        

    @property
    def cams_to_use_list(self):
        """ a list denoting which camera IDs to use.
         If both this and `num_cams` are set to default value (`0` and `None`, respectively), use all available cameras).
         """
        return self._cams_to_use_list

    @property
    def init_start_time(self):
        """Time that this object was initialized (UNIX Epoch time from `time.time_ns())

        Returns:
            float: time this object was created (UNIX Epoch time from `time.time_ns())
        """
        return self._init_start_time
    
    @property
    def each_cam_timestamps_unix_ns(self): #not really a property though?
        return self._each_cam_timestamps_unix_ns
    ###   
    ###   
    ###   ███    ███ ███████ ████████ ██   ██  ██████  ██████  ███████ 
    ###   ████  ████ ██         ██    ██   ██ ██    ██ ██   ██ ██      
    ###   ██ ████ ██ █████      ██    ███████ ██    ██ ██   ██ ███████ 
    ###   ██  ██  ██ ██         ██    ██   ██ ██    ██ ██   ██      ██ 
    ###   ██      ██ ███████    ██    ██   ██  ██████  ██████  ███████ 
    ###                                                                
    ###                                                                    
    def start(self):

        if self.cams_to_use_list is None:
            self.find_available_cameras()
        
        # pathos_pool = ProcessPool(nodes=1)
        # pathos_pool.map(self.start_multi_cam_thread_pool)
        # self.start_multi_cam_thread_pool() #threads are mostly fine, but I think processes will be better eventually? 
        self.start_multi_cam_process_pool() # as of 2021-10-26, this doesn't work :( 
        self.create_diagnostic_images()

    

    def find_available_cameras(self):
        """find available webcams and return IDs in list."""
        
        if self.num_cams == 0:
            num_cams_to_check = 20 #man, i HOPE you're out there trying to record from more than 20 cameras!!!
        else:
            num_cams_to_check = self.num_cams
        
        if platform.system() == 'Windows':
            capBackend = cv2.CAP_DSHOW
        else:
            capBackend = cv2.CAP_ANY

        self._cams_to_use_list = []
        for camNum in track(range(num_cams_to_check), description='[green]Finding available cameras...' ):
            cap =  cv2.VideoCapture(camNum, capBackend)
            if cap.isOpened():
                self._cams_to_use_list.append(camNum)
            cap.release()
        rich_console.print("Found cameras at ", self._cams_to_use_list)
        
        self.num_cams = len(self._cams_to_use_list)
        return self.cams_to_use_list
    ###  
    ###  
    ###  ███████ ████████  █████  ██████  ████████     ████████ ██   ██ ██████  ███████  █████  ██████      ██████   ██████   ██████  ██      
    ###  ██         ██    ██   ██ ██   ██    ██           ██    ██   ██ ██   ██ ██      ██   ██ ██   ██     ██   ██ ██    ██ ██    ██ ██      
    ###  ███████    ██    ███████ ██████     ██           ██    ███████ ██████  █████   ███████ ██   ██     ██████  ██    ██ ██    ██ ██      
    ###       ██    ██    ██   ██ ██   ██    ██           ██    ██   ██ ██   ██ ██      ██   ██ ██   ██     ██      ██    ██ ██    ██ ██      
    ###  ███████    ██    ██   ██ ██   ██    ██           ██    ██   ██ ██   ██ ███████ ██   ██ ██████      ██       ██████   ██████  ███████ 
    ###                                                                                                                                       
    ###                                                                                                                                       

    def start_multi_cam_thread_pool(self):
        """creates a threading.ThreadPoolExecutor that creates an `FMC_Camera` object for each camera in `self.cams_to_use_list` and sets each to trigger `run_in_thread` mode

            NOTE - Eventually I want these to be in different processes, I think?
        """
        self.cam_frame_queue = queue.Queue(maxsize=self.num_cams)
        self.multi_cam_image_queue = queue.Queue()

        self.exit_event = threading.Event()

        self._each_cam_timestamps_unix_ns = None
        self._cam_thread = [None] * self.num_cams

        rich_console.rule('starting multi_cam_thread_pool')

        while not self.exit_event.is_set():
            with concurrent.futures.ThreadPoolExecutor(max_workers=self.num_cams+1) as executor:
                self.barrier = threading.Barrier(self.num_cams, action=self.grab_incoming_cam_tuples)
                for  this_cam_num in self.cams_to_use_list: #i feel like there is some cooler way to do this using 'map' and 'zip' or whatever, but this feels easier on the ol' brain noggin lol
                    self._cam_thread[this_cam_num] = executor.submit(FMC_Camera, cam_num=this_cam_num, frame_queue = self.cam_frame_queue, barrier = self.barrier, exit_event = self.exit_event)
                executor.submit(self.show_multi_cam_cv2)

###   
###   
###   ███████ ████████  █████  ██████  ████████     ██████  ██████   ██████   ██████ ███████ ███████ ███████     ██████   ██████   ██████  ██      
###   ██         ██    ██   ██ ██   ██    ██        ██   ██ ██   ██ ██    ██ ██      ██      ██      ██          ██   ██ ██    ██ ██    ██ ██      
###   ███████    ██    ███████ ██████     ██        ██████  ██████  ██    ██ ██      █████   ███████ ███████     ██████  ██    ██ ██    ██ ██      
###        ██    ██    ██   ██ ██   ██    ██        ██      ██   ██ ██    ██ ██      ██           ██      ██     ██      ██    ██ ██    ██ ██      
###   ███████    ██    ██   ██ ██   ██    ██        ██      ██   ██  ██████   ██████ ███████ ███████ ███████     ██       ██████   ██████  ███████ 
###                                                                                                                                                
###                                                                                                                                                
###   
    def start_multi_cam_process_pool(self):
        """
        reates a ProcessPoolExecutor that creates an `FMC_Camera` object for each camera in `self.cams_to_use_list` and sets each to trigger `run_in_process`  mode (which just passes to `run_in_thread` mode)

        """
        mrManager = pathos_mp_helper.Manager()
        self.cam_frame_queue = mrManager.Queue(maxsize=self.num_cams)
        self.multi_cam_image_queue = mrManager.Queue()
        self.barrier = mrManager.Barrier(self.num_cams+1) #each cam + one more for the frame grabber
        self.exit_event = mrManager.Event()

        self._each_cam_timestamps_unix_ns = None
        self._cam_process = [None] * self.num_cams

        rich_console.rule('starting multi_cam_process_pool')

        in_frame_q_list = [self.cam_frame_queue]*self.num_cams
        in_barrier_list = [self.barrier]*self.num_cams
        in_exit_event_list = [self.exit_event]*self.num_cams
        # nsole_list = [self.rich_console]*self.num_cams


        num_jobs = self.num_cams
        with pathos_mp.ProcessingPool(num_jobs) as pp_pool: #lol
            pp_pool.amap(
                        FMC_Camera, 
                        tuple(self.cams_to_use_list), 
                        tuple(in_frame_q_list), 
                        tuple(in_barrier_list), 
                        tuple(in_exit_event_list),
                        )
            incoming_frame_grabber_thread = threading.Thread(target=self.grab_incoming_cam_tuples, name='Incoming Frame Grabber')
            incoming_frame_grabber_thread.start()
            self.show_multi_cam_cv2()



    ###  
    ###  
    ###   ██████  ██████   █████  ██████      ██ ███    ██  ██████  ██████  ███    ███ ██ ███    ██  ██████      ████████ ██    ██ ██████  ██      ███████ ███████ 
    ###  ██       ██   ██ ██   ██ ██   ██     ██ ████   ██ ██      ██    ██ ████  ████ ██ ████   ██ ██              ██    ██    ██ ██   ██ ██      ██      ██      
    ###  ██   ███ ██████  ███████ ██████      ██ ██ ██  ██ ██      ██    ██ ██ ████ ██ ██ ██ ██  ██ ██   ███        ██    ██    ██ ██████  ██      █████   ███████ 
    ###  ██    ██ ██   ██ ██   ██ ██   ██     ██ ██  ██ ██ ██      ██    ██ ██  ██  ██ ██ ██  ██ ██ ██    ██        ██    ██    ██ ██      ██      ██           ██ 
    ###   ██████  ██   ██ ██   ██ ██████      ██ ██   ████  ██████  ██████  ██      ██ ██ ██   ████  ██████         ██     ██████  ██      ███████ ███████ ███████ 
    ###                                                                                                                                                            
    ###                                                                                                                                                            


    def grab_incoming_cam_tuples(self):
        """grab incoming cam_tuples, put them into a multi_cam_tuple and stuff that into the multi_cam_tuple_queue"""

        # self.show_multi_cam_cv2()

        while not self.exit_event.is_set():
            self.barrier.wait() #wait until each camera has grabbed an image and tagged their `barrier.wait()` signals

            these_images_list = [None]*self.num_cams #empty list of size (numCam)
            these_timestamps = np.ndarray(self.num_cams)

            for cam_num in range(self.num_cams):
                this_cam_image_timestamp_tuple = self.cam_frame_queue.get()
                this_cam_num = this_cam_image_timestamp_tuple[0]
                these_images_list[this_cam_num] = this_cam_image_timestamp_tuple[1]
                these_timestamps[this_cam_num] = this_cam_image_timestamp_tuple[2]
            
            if self._each_cam_timestamps_unix_ns is None:
                self._each_cam_timestamps_unix_ns =  these_timestamps
            else:
                self._each_cam_timestamps_unix_ns =  np.vstack((self._each_cam_timestamps_unix_ns, these_timestamps)) #result will be numpy array with `num_frames` rows and `num_cams` columns


            multi_cam_image = np.hstack(these_images_list)  #create multiFrame_image by stitching together incoming camera images
            self.multi_cam_image_queue.put(multi_cam_image)
            # rich_console.log('Created a multi_cam_image - queue size: {}'.format(self.multi_cam_image_queue.qsize()))
    
    ###                  
    ###  
    ###  ███████ ██   ██  ██████  ██     ██     ██    ██ ██ ██████  
    ###  ██      ██   ██ ██    ██ ██     ██     ██    ██ ██ ██   ██ 
    ###  ███████ ███████ ██    ██ ██  █  ██     ██    ██ ██ ██   ██ 
    ###       ██ ██   ██ ██    ██ ██ ███ ██      ██  ██  ██ ██   ██ 
    ###  ███████ ██   ██  ██████   ███ ███        ████   ██ ██████  
    ###                                                             
    ###                                                             
    
    def show_multi_cam_cv2(self):
        """display multi_cam_image using `cv2.imshow` and maybe save it to mp4, who knows?
        """
        rich_console.rule('Launching Multi Cam Viewer')
        while not self.exit_event.is_set():   
            if not self.multi_cam_image_queue.empty():
                multi_cam_image = self.multi_cam_image_queue.get()

                if self._save_to_mp4:
                    if self._output_video_object is None:
                        self.initialize_video(multi_cam_image)
                    self._output_video_object.write(multi_cam_image)

                cv2.imshow(self._rec_name, multi_cam_image)
                key = cv2.waitKey(1)

                if key == 27:  # exit on ESC                        
                    self.exit_event.set()
                
                if cv2.getWindowProperty(self._rec_name, cv2.WND_PROP_VISIBLE) < 1: #break loop if window closed
                    self.exit_event.set()   

        cv2.destroyAllWindows()
        if self._save_to_mp4:
            self._output_video_object.release()
        rich_console.rule('Shutting down MultiCamera Viewer')
        self.exit_event.set() #send the 'Exit' signal to everyone.
    
    ###  
    ###          
    ###  ██ ███    ██ ██ ████████      ██████  ██    ██ ████████     ██    ██ ██ ██████  
    ###  ██ ████   ██ ██    ██        ██    ██ ██    ██    ██        ██    ██ ██ ██   ██ 
    ###  ██ ██ ██  ██ ██    ██        ██    ██ ██    ██    ██        ██    ██ ██ ██   ██ 
    ###  ██ ██  ██ ██ ██    ██        ██    ██ ██    ██    ██         ██  ██  ██ ██   ██ 
    ###  ██ ██   ████ ██    ██         ██████   ██████     ██          ████   ██ ██████  
    ###                                                                                  
    ###                                                                                  

    def initialize_video(self, multi_cam_image):
        """use  the multi_cam_image to initialize the video
            multicam image will be of size  [cam_resolution_width*num_cams by cam_resolution_height]
        Args:
            multi_cam_image ([type]): [description]
        """
        #create video save object
        self._multi_cam_image_height, self._multi_cam_image_width, channels = multi_cam_image.shape

        self.multi_cam_image_size = ( self._multi_cam_image_width, self._multi_cam_image_height)
        fourcc = cv2.VideoWriter_fourcc(*'DIVX')
        self._outputVid_fileName = str(self._save_path / self._rec_name) + '_outVid.mp4'
        fps = 30 #this is a bad and stupid guess, but it's actually kinda tricky to guess what is the right thing to put here. I'll fix this later and expect videos to be a bit faster than usual :(
        self._output_video_object = cv2.VideoWriter(self._outputVid_fileName, fourcc, fps, self.multi_cam_image_size)
    ##   
    ##   
    ##   ██████  ██       ██████  ████████     ████████ ██ ███    ███ ███████ ███████ ████████  █████  ███    ███ ██████  ███████ 
    ##   ██   ██ ██      ██    ██    ██           ██    ██ ████  ████ ██      ██         ██    ██   ██ ████  ████ ██   ██ ██      
    ##   ██████  ██      ██    ██    ██           ██    ██ ██ ████ ██ █████   ███████    ██    ███████ ██ ████ ██ ██████  ███████ 
    ##   ██      ██      ██    ██    ██           ██    ██ ██  ██  ██ ██           ██    ██    ██   ██ ██  ██  ██ ██           ██ 
    ##   ██      ███████  ██████     ██           ██    ██ ██      ██ ███████ ███████    ██    ██   ██ ██      ██ ██      ███████ 
    ##                                                                                                                            
    ##                                                                                                                            

    def create_diagnostic_images(self):
        """plot some diagnostics to assess quality of camera sync"""
        try:
            camTimestamps = (self.each_cam_timestamps_unix_ns-self.init_start_time)/1e9 #subtract start time and convert to sec
            
            npy_save_path = self._save_path /'multi_cam_timestamps_frameNum_camNum.npy'
            np.save(str(npy_save_path), camTimestamps)
            rich_console.log('saving timestamp data to - ' + str(npy_save_path) )
            
            meanMultiFrameTimestamp = np.mean(camTimestamps, axis=1)
            meanMultiFrameTimespan = np.max(camTimestamps, axis=1) - np.min(camTimestamps, axis=1) #what was the timespan covered by each frame
            plt.ion()
            fig = plt.figure(figsize=(18,10))
            max_frame_duration = .1
            ax1  = plt.subplot(231, title='Camera Frame Timestamp vs Frame#', xlabel='Frame#', ylabel='Timestamp (sec)')
            ax2  = plt.subplot(232, ylim=(0,max_frame_duration), title='Camera Frame Duration Trace', xlabel='Frame#', ylabel='Duration (sec)')
            ax3  = plt.subplot(233, xlim=(0,max_frame_duration), title='Camera Frame Duration Histogram (count)', xlabel='Duration(s, 1ms bins)', ylabel='Probability')
            ax4  = plt.subplot(234,  title='MuliFrame Timestamp vs Frame#', xlabel='Frame#', ylabel='Timestamp (sec)')
            ax5  = plt.subplot(235,  ylim=(0,max_frame_duration), title='Multi Frame Duration/Span Trace', xlabel='Frame#', ylabel='Duration (sec)')
            ax6  = plt.subplot(236, xlim=(0,max_frame_duration), title='MultiFrame Duration Histogram (count)', xlabel='Duration(s, 1ms bins)', ylabel='Probability')

            for camNum in range(self.num_cams):
                thisCamTimestamps = camTimestamps[:,camNum]
                ax1.plot(thisCamTimestamps, label='Camera#'+str(camNum))
                ax1.legend()
                ax2.plot(np.diff(thisCamTimestamps),'.')    
                ax3.hist(np.diff(thisCamTimestamps), bins=np.arange(0,max_frame_duration,.001), alpha=0.5)

            ax4.plot(meanMultiFrameTimestamp, color='darkslategrey', label='MultiFrame'+str(camNum))
            ax5.plot(np.diff(meanMultiFrameTimestamp),'.',color='darkslategrey', label='Frame Duration')    
            ax5.plot(meanMultiFrameTimespan, '.', color='orangered', label='Frame TimeSpan')    
            ax5.legend()
            ax6.hist(np.diff(meanMultiFrameTimestamp), bins=np.arange(0,max_frame_duration,.001), density=True, alpha=0.5, color='darkslategrey', label='Frame Duration')
            ax6.hist(np.diff(meanMultiFrameTimespan), bins=np.arange(0,max_frame_duration,.001), density=True, alpha=0.5, color='orangered', label='Frame Timespan')
            ax5.legend()
        
            fig_save_path = self._save_path / 'recording_diagnostics.png'
            plt.savefig(str(fig_save_path))  
            rich_console.log('Saving diagnostic figure to - '+str(fig_save_path))
            plt.show()
            plt.waitforbuttonpress()
        except:
            rich_console.print_exception()
        f=9
    
    ### 
    ### ███████ ████████  ██████              
    ### ██         ██    ██                   
    ### █████      ██    ██                   
    ### ██         ██    ██                   
    ### ███████    ██     ██████     ██ ██ ██ 
    ###                                       
    
    def __enter__(self):
        """Context manager -  No need to do anything special on start"""
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """Context manager -  on exit, set `self.exit_event` to activate shutdown sequence """
        return self
###   
###               
###   ██ ███████                     ███    ███  █████  ██ ███    ██                 
###   ██ ██                          ████  ████ ██   ██ ██ ████   ██                 
###   ██ █████                       ██ ████ ██ ███████ ██ ██ ██  ██                 
###   ██ ██                          ██  ██  ██ ██   ██ ██ ██  ██ ██                 
###   ██ ██          ███████ ███████ ██      ██ ██   ██ ██ ██   ████ ███████ ███████ 
###                                                                                  
###                                                                                  


if __name__ == '__main__':
    pathos_mp_helper.freeze_support()
    console = Console() #create rich console to catch and print exceptions
    try:
        fmc_multi_cam_recorder = FMC_MultiCameraRecorder()
        fmc_multi_cam_recorder.start()        
                                 
    except Exception:
        console.print_exception()


